# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

import datetime

from django.conf import settings
from django.contrib.auth.hashers import make_password
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.operations import BtreeGistExtension
from django.core.serializers.json import DjangoJSONEncoder
from django.core.validators import MaxLengthValidator, MinLengthValidator
from django.db import migrations, models
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps
from django.db.models.expressions import CombinedExpression
from django.db.models.fields.json import KeyTextTransform
from django.utils import timezone
from pgtrigger.compiler import Trigger, UpsertTriggerSql

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    TaskTypes,
)
from debusine.assets.models import AssetCategory
from debusine.db.constraints import JsonDataUniqueConstraint
from debusine.db.models import DEFAULT_FILE_STORE_NAME, SYSTEM_USER_NAME
from debusine.db.models.auth import UserManager, validate_group_name
from debusine.db.models.collections import validate_collection_name
from debusine.db.models.files import _FileStoreBackendChoices
from debusine.db.models.scopes import validate_scope_name
from debusine.db.models.work_requests import validate_notification_channel_name
from debusine.db.models.workspaces import validate_workspace_name
from debusine.tasks.models import WorkerType

# We only create the collections that were defined as singletons at the time
# of this migration.  If we add new singleton categories and want to create
# them in the default workspace as well, then that requires a new migration.
singleton_collection_categories = (
    CollectionCategory.PACKAGE_BUILD_LOGS,
    CollectionCategory.TASK_HISTORY,
)


def create_default_objects(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    Collection = apps.get_model("db", "Collection")
    FileStore = apps.get_model("db", "FileStore")
    Scope = apps.get_model("db", "Scope")
    Workspace = apps.get_model("db", "Workspace")
    User = apps.get_model("db", "User")

    default_file_store = FileStore.objects.create(
        name=DEFAULT_FILE_STORE_NAME,
        backend=_FileStoreBackendChoices.LOCAL,
        configuration={},
    )
    default_scope = Scope.objects.create(
        name=settings.DEBUSINE_DEFAULT_SCOPE,
        label=settings.DEBUSINE_DEFAULT_SCOPE.capitalize(),
    )
    default_scope.file_stores.set(
        [default_file_store],
        through_defaults={"upload_priority": 100, "download_priority": 100},
    )
    default_workspace = Workspace.objects.create(
        scope=default_scope,
        name=settings.DEBUSINE_DEFAULT_WORKSPACE,
        public=True,
    )

    unusable_password = make_password(None)
    User.objects.create(
        username=SYSTEM_USER_NAME,
        password=unusable_password,
        first_name="Debusine",
        last_name="System",
        is_system=True,
    )

    for category in singleton_collection_categories:
        Collection.objects.create(
            name="_", category=category, workspace=default_workspace
        )
    Collection.objects.create(
        name="default",
        category=CollectionCategory.TASK_CONFIGURATION,
        workspace=default_workspace,
    )


class Migration(migrations.Migration):
    replaces = [
        ("db", "0001_initial"),
        ("db", "0002_create_default_store"),
        ("db", "0003_create_workspace"),
        ("db", "0004_add_field_user_to_token"),
        ("db", "0005_workrequest_created_by"),
        ("db", "0006_alter_workrequest_created_by"),
        ("db", "0007_notificationchannel"),
        ("db", "0008_artifact_created_by"),
        ("db", "0009_alter_token_comment"),
        ("db", "0010_identity"),
        ("db", "0011_auto_20231024_1213"),
        ("db", "0012_workspace_default_expiration_delay"),
        ("db", "0013_expiration_delay"),
        ("db", "0014_artifact_underscore_names"),
        ("db", "0015_alter_fileinartifact_file_non_null"),
        ("db", "0016_worker_internal"),
        ("db", "0017_collections"),
        ("db", "0018_workrequest_worker_blank"),
        ("db", "0019_autopkgtest_task_data"),
        ("db", "0020_alter_filestore_backend"),
        ("db", "0021_rename_constraints"),
        (
            "db",
            "0022_collectionitem_db_collectionitem_unique_debian_environment",
        ),
        ("db", "0023_token_hash_initial"),
        ("db", "0024_token_hash_data"),
        ("db", "0025_token_hash_final"),
        ("db", "0026_suite_collection"),
        ("db", "0027_workrequest_task_type"),
        ("db", "0028_workrequest_changes_for_workflows"),
        ("db", "0029_workflowtemplate"),
        ("db", "0030_move_task_type"),
        ("db", "0031_workrequest_priorities"),
        ("db", "0032_workrequest_manage_priorities_permission"),
        ("db", "0033_collectionitem_tighten_unique_active_name"),
        ("db", "0034_workrequest_pydantic"),
        ("db", "0035_autopkgtest_extra_environment"),
        ("db", "0036_workrequest_fix_task_type"),
        ("db", "0037_workrequest_internal_collection"),
        ("db", "0038_alter_workrequest_parent"),
        ("db", "0039_rename_environment_id"),
        ("db", "0040_blhc_lookup"),
        ("db", "0041_piuparts_lookup"),
        ("db", "0042_user_is_system"),
        ("db", "0043_create_system_user"),
        ("db", "0044_collection_add_retains_artifacts"),
        ("db", "0045_autopkgtest_lookup"),
        ("db", "0046_lintian_lookup"),
        ("db", "0047_sbuild_lookup"),
        ("db", "0048_alter_collection_retains_artifacts"),
        ("db", "0049_update_suite_lintian_collection_lookup"),
        ("db", "0050_collection_workflow_initial"),
        ("db", "0051_collection_workflow_data"),
        ("db", "0052_collection_workflow_final"),
        ("db", "0053_debian_environments_typo"),
        ("db", "0054_alter_workrequest_options"),
        ("db", "0055_artifact_created_by_work_request_set_null"),
        ("db", "0056_debian_environments_backend"),
        ("db", "0057_workrequest_expiration_delay"),
        ("db", "0058_workrequest_dynamic_task_data"),
        ("db", "0059_collection_workspace_namespace"),
        ("db", "0060_workspace_chain"),
        ("db", "0061_workrequest_supersedes"),
        ("db", "0062_rename_internalnoop"),
        ("db", "0063_worker_worker_type_initial"),
        ("db", "0064_worker_worker_type_data"),
        ("db", "0065_worker_worker_type_final"),
        ("db", "0066_signing_worker_type"),
        ("db", "0067_alter_filestore_backend"),
        ("db", "0068_collection_workflow_onetoone"),
        ("db", "0069_collectionitem_parent_category_initial"),
        ("db", "0070_collectionitem_parent_category_data"),
        ("db", "0071_collectionitem_parent_category_final"),
        ("db", "0072_worker_last_seen_at"),
        ("db", "0073_token_last_seen_at"),
        ("db", "0074_fileinartifact_complete_initial"),
        ("db", "0075_fileinartifact_complete_data"),
        ("db", "0076_fileinartifact_complete_final"),
        ("db", "0077_wait_task_type"),
        ("db", "0078_collectionitem_add_created_removed_by_workflow"),
        ("db", "0079_sbuild_binnmu_build_components"),
        ("db", "0080_scope"),
        ("db", "0081_create_fallback_scope"),
        ("db", "0082_workspace_add_scope"),
        ("db", "0083_set_fallback_scope"),
        ("db", "0084_workspace_namespace_by_scope"),
        ("db", "0085_scope_name_validation"),
        ("db", "0086_add_scoped_group"),
        ("db", "0087_django_json_encoder"),
        ("db", "0088_roles_for_scopes"),
        ("db", "0089_group_name_validation"),
        ("db", "0090_workspacerole_and_more"),
        ("db", "0091_sign_multiple_unsigned"),
        ("db", "0092_default_workspace_public"),
        ("db", "0093_uniform_scoperole_resource_reference"),
        ("db", "0094_collection_db_collection_name_not_reserved"),
        ("db", "0095_create_singleton_collections"),
        ("db", "0096_workspace_name_validation"),
        ("db", "0097_recreate_json_data_unique_constraint"),
        ("db", "0098_artifact_original_artifact"),
        ("db", "0099_delete_dynamic_data_workflows"),
        ("db", "0100_name_validation"),
        ("db", "0101_collectionitem_suite_indexes"),
        ("db", "0102_collection_related_name"),
        ("db", "0103_cascade_deletion_on_role_assignments"),
        ("db", "0104_workspacerole_contributor"),
        ("db", "0105_wait_needs_input"),
        ("db", "0106_workrequest_workflow_template"),
        ("db", "0107_add_scope_label_icon"),
        ("db", "0108_fill_scope_label"),
        ("db", "0109_scope_label_required_and_unique"),
        ("db", "0110_workrequest_workflow_last_activity_at"),
        ("db", "0111_workrequest_workflow_runtime_status"),
        ("db", "0112_workrequest_autopkgtest_extra_repositories"),
        ("db", "0113_workrequest_output_data_json"),
        ("db", "0114_user_manager_with_permissions"),
        ("db", "0115_workrequest_workflow_runtime_status_use_value"),
        ("db", "0116_assets"),
        ("db", "0117_signing_work_request_fingerprints"),
        ("db", "0118_signing_key_assets"),
        ("db", "0119_scope_file_stores_initial"),
        ("db", "0120_scope_file_stores_data"),
        ("db", "0121_remove_workspace_file_stores"),
        ("db", "0122_create_default_task_configuration_collections"),
        ("db", "0123_asset_cloud_provider_account"),
        (
            "db",
            "0124_drop_category_from_db_asset_unique_signing_key_fingerprints",
        ),
        ("db", "0125_add_group_ephemeral"),
        ("db", "0126_filestore_instance_wide"),
        ("db", "0127_workspace_add_expiry_fields"),
        ("db", "0128_filestore_add_size_columns"),
        ("db", "0129_filestore_total_size"),
        ("db", "0130_filestore_total_size_triggers"),
        ("db", "0131_workrequest_add_configured_task_data_and_version"),
        ("db", "0132_filestore_size_columns_biginteger"),
        ("db", "0133_filestore_provider_account_alter_filestore_backend"),
        ("db", "0134_filestoreinscope_db_filestoreinscope_consistent_populate"),
        ("db", "0135_workerpools"),
        ("db", "0136_add_groupmembership"),
        ("db", "0137_add_groupmembership_role"),
        ("db", "0138_worker_activation_token"),
        ("db", "0139_groupmembership_unique_role"),
        ("db", "0140_groupauditlog"),
        ("db", "0141_add_clientenroll"),
        ("db", "0142_workrequest_workflows_need_update"),
    ]

    initial = True

    dependencies = [("auth", "0012_alter_user_first_name_max_length")]

    operations = [
        BtreeGistExtension(),
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "password",
                    models.CharField(max_length=128, verbose_name="password"),
                ),
                (
                    "last_login",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="last login"
                    ),
                ),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to each of their groups.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Specific permissions for this user.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.permission",
                        verbose_name="user permissions",
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        error_messages={
                            "unique": "A user with that username already exists."
                        },
                        help_text="Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.",
                        max_length=150,
                        unique=True,
                        validators=[UnicodeUsernameValidator()],
                        verbose_name="username",
                    ),
                ),
                (
                    "first_name",
                    models.CharField(
                        blank=True, max_length=150, verbose_name="first name"
                    ),
                ),
                (
                    "last_name",
                    models.CharField(
                        blank=True, max_length=150, verbose_name="last name"
                    ),
                ),
                (
                    "email",
                    models.EmailField(
                        blank=True, max_length=254, verbose_name="email address"
                    ),
                ),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                (
                    "date_joined",
                    models.DateTimeField(
                        default=timezone.now, verbose_name="date joined"
                    ),
                ),
                ("is_system", models.BooleanField(default=False)),
            ],
            options={
                "verbose_name": "user",
                "verbose_name_plural": "users",
                "abstract": False,
                "constraints": [
                    models.CheckConstraint(
                        check=models.Q(
                            ("is_system", False), ("email", ""), _connector="OR"
                        ),
                        name="db_user_non_system_email",
                    ),
                    models.UniqueConstraint(
                        condition=models.Q(("is_system", False)),
                        fields=("email",),
                        name="db_user_unique_email",
                    ),
                ],
            },
            managers=[("objects", UserManager())],
        ),
        migrations.CreateModel(
            name="ClientEnroll",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("nonce", models.CharField(max_length=64, unique=True)),
                ("payload", models.JSONField(default=dict)),
            ],
        ),
        migrations.CreateModel(
            name="File",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "sha256",
                    models.BinaryField(
                        help_text="sha256 of the file", max_length=4
                    ),
                ),
                (
                    "size",
                    models.PositiveBigIntegerField(
                        help_text="Size in bytes of the file"
                    ),
                ),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(
                        fields=("sha256", "size"),
                        name="db_file_unique_sha256_size",
                    ),
                    models.CheckConstraint(
                        check=models.Q(("sha256", b""), _negated=True),
                        name="db_file_sha256_not_empty",
                    ),
                ]
            },
        ),
        migrations.CreateModel(
            name="FileStore",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255, unique=True)),
                (
                    "backend",
                    models.CharField(
                        choices=[
                            ("Local", "Local"),
                            ("ExternalDebianSuite", "ExternalDebianSuite"),
                            ("Memory", "Memory"),
                            ("S3", "S3"),
                        ],
                        max_length=255,
                    ),
                ),
                ("configuration", models.JSONField(blank=True, default=dict)),
                ("instance_wide", models.BooleanField(default=True)),
                (
                    "total_size",
                    models.BigIntegerField(default=0, editable=False),
                ),
                (
                    "soft_max_size",
                    models.BigIntegerField(blank=True, null=True),
                ),
                ("max_size", models.BigIntegerField(blank=True, null=True)),
            ],
            options={
                "triggers": [
                    Trigger(
                        name="db_filestore_sync_instance_wide",
                        sql=UpsertTriggerSql(
                            func="UPDATE db_filestoreinscope SET file_store_instance_wide = NEW.instance_wide WHERE db_filestoreinscope.file_store_id = NEW.id AND db_filestoreinscope.file_store_instance_wide != NEW.instance_wide; RETURN NEW;",
                            hash="7dfd74c5dbb675f55873e1fa249285a172869cf3",
                            operation="INSERT OR UPDATE",
                            pgid="pgtrigger_db_filestore_sync_instance_wide_e2fab",
                            table="db_filestore",
                            when="AFTER",
                        ),
                    )
                ],
                "constraints": [
                    models.CheckConstraint(
                        check=models.Q(
                            ("soft_max_size__gt", models.F("max_size")),
                            _negated=True,
                        ),
                        name="db_filestore_consistent_max_sizes",
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="Scope",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="internal name for the scope",
                        max_length=255,
                        unique=True,
                        validators=[validate_scope_name],
                    ),
                ),
                (
                    "label",
                    models.CharField(
                        help_text="User-visible name for the scope",
                        max_length=255,
                        unique=True,
                    ),
                ),
                (
                    "icon",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Optional user-visible icon, resolved via ``{% static %}`` in templates",
                        max_length=255,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="FileStoreInScope",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "scope",
                    models.ForeignKey(on_delete=models.PROTECT, to="db.scope"),
                ),
                (
                    "file_store",
                    models.ForeignKey(
                        on_delete=models.PROTECT, to="db.filestore"
                    ),
                ),
                (
                    "_file_store_instance_wide",
                    models.BooleanField(
                        db_column="file_store_instance_wide",
                        default=True,
                        editable=False,
                        help_text="Synced from FileStore.instance_wide; do not update directly",
                    ),
                ),
                ("upload_priority", models.IntegerField(blank=True, null=True)),
                (
                    "download_priority",
                    models.IntegerField(blank=True, null=True),
                ),
                ("drain", models.BooleanField(default=False)),
                ("drain_to", models.TextField(blank=True, null=True)),
                ("populate", models.BooleanField(default=False)),
                ("read_only", models.BooleanField(default=False)),
                ("soft_max_size", models.IntegerField(blank=True, null=True)),
                ("write_only", models.BooleanField(default=False)),
            ],
            options={
                "triggers": [
                    Trigger(
                        name="db_filestoreinscope_sync_instance_wide",
                        sql=UpsertTriggerSql(
                            func="NEW.file_store_instance_wide = (SELECT db_filestore.instance_wide FROM db_filestore WHERE db_filestore.id = NEW.file_store_id); RETURN NEW;",
                            hash="04b1ed9053b9e9ea2cf78d10850742c925eeae66",
                            operation="INSERT OR UPDATE",
                            pgid="pgtrigger_db_filestoreinscope_sync_instance_wide_c82e6",
                            table="db_filestoreinscope",
                            when="BEFORE",
                        ),
                    )
                ],
                "constraints": [
                    models.UniqueConstraint(
                        fields=("scope", "file_store"),
                        name="db_filestoreinscope_unique_scope_file_store",
                    ),
                    models.UniqueConstraint(
                        condition=models.Q(
                            ("_file_store_instance_wide", False)
                        ),
                        fields=("file_store",),
                        name="db_filestoreinscope_unique_file_store_not_instance_wide",
                    ),
                    models.CheckConstraint(
                        check=models.Q(
                            ("populate", False),
                            models.Q(("drain", False), ("read_only", False)),
                            _connector="OR",
                        ),
                        name="db_filestoreinscope_consistent_populate",
                    ),
                ],
            },
        ),
        migrations.CreateModel(
            name="Group",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        max_length=255, validators=[validate_group_name]
                    ),
                ),
                (
                    "scope",
                    models.ForeignKey(on_delete=models.PROTECT, to="db.scope"),
                ),
                (
                    "ephemeral",
                    models.BooleanField(
                        default=False,
                        help_text="remove the group if it has no roles assigned",
                    ),
                ),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(
                        fields=("name", "scope"),
                        name="db_group_unique_name_scope",
                    )
                ]
            },
        ),
        migrations.CreateModel(
            name="GroupAuditLog",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("changes", models.JSONField()),
                (
                    "actor",
                    models.ForeignKey(
                        null=True,
                        on_delete=models.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="audit_log",
                        to="db.group",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="GroupMembership",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="membership",
                        to="db.group",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="group_memberships",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "role",
                    models.CharField(
                        choices=[("admin", "Admin"), ("member", "Member")],
                        default="member",
                        max_length=16,
                    ),
                ),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(
                        fields=("group", "user"),
                        name="db_groupmembership_unique_group_user",
                    )
                ]
            },
        ),
        migrations.CreateModel(
            name="ScopeRole",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "resource",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="roles",
                        to="db.scope",
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="scope_roles",
                        to="db.group",
                    ),
                ),
                (
                    "role",
                    models.CharField(
                        choices=[("owner", "Owner")], max_length=16
                    ),
                ),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(
                        fields=("resource", "group", "role"),
                        name="db_scoperole_unique_resource_group_role",
                    )
                ]
            },
        ),
        migrations.CreateModel(
            name="Workspace",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        max_length=255, validators=[validate_workspace_name]
                    ),
                ),
                ("public", models.BooleanField(default=False)),
                (
                    "default_expiration_delay",
                    models.DurationField(
                        default=datetime.timedelta(0),
                        help_text="minimal time that a new artifact is kept in the workspace before being expired",
                    ),
                ),
                (
                    "scope",
                    models.ForeignKey(
                        on_delete=models.PROTECT,
                        related_name="workspaces",
                        to="db.scope",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="time the workspace was created",
                    ),
                ),
                (
                    "expiration_delay",
                    models.DurationField(
                        blank=True,
                        help_text="if set, time since the last task completion timeafter which the workspace can be deleted",
                        null=True,
                    ),
                ),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(
                        fields=("scope", "name"),
                        name="db_workspace_unique_scope_name",
                    )
                ]
            },
        ),
        migrations.CreateModel(
            name="WorkspaceChain",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "child",
                    models.ForeignKey(
                        help_text="Workspace that falls back on `parent` for lookups",
                        on_delete=models.CASCADE,
                        related_name="chain_parents",
                        to="db.workspace",
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        help_text="Workspace to be looked up if lookup in `child` fails",
                        on_delete=models.CASCADE,
                        related_name="chain_children",
                        to="db.workspace",
                    ),
                ),
                (
                    "order",
                    models.IntegerField(
                        help_text="Lookup order of this element in the chain"
                    ),
                ),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(
                        fields=("child", "parent"),
                        name="db_workspacechain_unique_child_parent",
                    ),
                    models.UniqueConstraint(
                        fields=("child", "order"),
                        name="db_workspacechain_unique_child_order",
                    ),
                ]
            },
        ),
        migrations.CreateModel(
            name="WorkspaceRole",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "resource",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="roles",
                        to="db.workspace",
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="workspace_roles",
                        to="db.group",
                    ),
                ),
                (
                    "role",
                    models.CharField(
                        choices=[
                            ("owner", "Owner"),
                            ("contributor", "Contributor"),
                        ],
                        max_length=16,
                    ),
                ),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(
                        fields=("resource", "group", "role"),
                        name="db_workspacerole_unique_resource_group_role",
                    )
                ]
            },
        ),
        migrations.CreateModel(
            name="Asset",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "category",
                    models.CharField(
                        choices=[
                            (
                                "debusine:cloud-provider-account",
                                "debusine:cloud-provider-account",
                            ),
                            ("debusine:signing-key", "debusine:signing-key"),
                        ],
                        max_length=255,
                    ),
                ),
                (
                    "workspace",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=models.PROTECT,
                        to="db.workspace",
                    ),
                ),
                ("data", models.JSONField(default=dict)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=models.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "constraints": [
                    models.CheckConstraint(
                        check=models.Q(
                            models.Q(
                                ("category", AssetCategory["SIGNING_KEY"]),
                                _negated=True,
                            ),
                            ("workspace__isnull", False),
                            _connector="OR",
                        ),
                        name="db_asset_workspace_not_null",
                    )
                ]
            },
        ),
        migrations.CreateModel(
            name="AssetUsage",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "asset",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="usage",
                        to="db.asset",
                    ),
                ),
                (
                    "workspace",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="asset_usage",
                        to="db.workspace",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AssetUsageRole",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "role",
                    models.CharField(
                        choices=[("signer", "Signer")], max_length=16
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="asset_usage_roles",
                        to="db.group",
                    ),
                ),
                (
                    "resource",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="roles",
                        to="db.assetusage",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AssetRole",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "role",
                    models.CharField(
                        choices=[("owner", "Owner")], max_length=16
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="asset_roles",
                        to="db.group",
                    ),
                ),
                (
                    "resource",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="roles",
                        to="db.asset",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Artifact",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("category", models.CharField(max_length=255)),
                (
                    "workspace",
                    models.ForeignKey(
                        on_delete=models.PROTECT, to="db.workspace"
                    ),
                ),
                ("data", models.JSONField(blank=True, default=dict)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "expiration_delay",
                    models.DurationField(blank=True, null=True),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=models.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "original_artifact",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=models.SET_NULL,
                        to="db.artifact",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ArtifactRelation",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "artifact",
                    models.ForeignKey(
                        on_delete=models.PROTECT,
                        related_name="relations",
                        to="db.artifact",
                    ),
                ),
                (
                    "target",
                    models.ForeignKey(
                        on_delete=models.PROTECT,
                        related_name="targeted_by",
                        to="db.artifact",
                    ),
                ),
                (
                    "type",
                    models.CharField(
                        choices=[
                            ("extends", "Extends"),
                            ("relates-to", "Relates to"),
                            ("built-using", "Built using"),
                        ],
                        max_length=11,
                    ),
                ),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(
                        fields=("artifact", "target", "type"),
                        name="db_artifactrelation_unique_artifact_target_type",
                    )
                ]
            },
        ),
        migrations.CreateModel(
            name="Collection",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        max_length=255, validators=[validate_collection_name]
                    ),
                ),
                ("category", models.CharField(max_length=255)),
                (
                    "full_history_retention_period",
                    models.DurationField(blank=True, null=True),
                ),
                (
                    "metadata_only_retention_period",
                    models.DurationField(blank=True, null=True),
                ),
                (
                    "workspace",
                    models.ForeignKey(
                        on_delete=models.PROTECT,
                        related_name="collections",
                        to="db.workspace",
                    ),
                ),
                (
                    "retains_artifacts",
                    models.CharField(
                        choices=[
                            ("never", "Never"),
                            ("workflow", "While workflow is running"),
                            ("always", "Always"),
                        ],
                        default="always",
                        max_length=8,
                    ),
                ),
                ("data", models.JSONField(blank=True, default=dict)),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(
                        fields=("name", "category", "workspace"),
                        name="db_collection_unique_name_category_workspace",
                    ),
                    models.CheckConstraint(
                        check=models.Q(("name", ""), _negated=True),
                        name="db_collection_name_not_empty",
                    ),
                    models.CheckConstraint(
                        check=models.Q(
                            models.Q(
                                models.Q(
                                    (
                                        "category__in",
                                        {
                                            CollectionCategory[
                                                "PACKAGE_BUILD_LOGS"
                                            ],
                                            CollectionCategory["TASK_HISTORY"],
                                        },
                                    ),
                                    _negated=True,
                                ),
                                models.Q(
                                    ("name__startswith", "_"), _negated=True
                                ),
                            ),
                            models.Q(
                                (
                                    "category__in",
                                    {
                                        CollectionCategory[
                                            "PACKAGE_BUILD_LOGS"
                                        ],
                                        CollectionCategory["TASK_HISTORY"],
                                    },
                                ),
                                ("name", "_"),
                            ),
                            _connector="OR",
                        ),
                        name="db_collection_name_not_reserved",
                    ),
                    models.CheckConstraint(
                        check=models.Q(("category", ""), _negated=True),
                        name="db_collection_category_not_empty",
                    ),
                ]
            },
        ),
        migrations.CreateModel(
            name="CollectionItem",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                (
                    "child_type",
                    models.CharField(
                        choices=[
                            ("b", "Bare"),
                            ("a", "Artifact"),
                            ("c", "Collection"),
                        ],
                        max_length=1,
                    ),
                ),
                ("category", models.CharField(max_length=255)),
                (
                    "parent_collection",
                    models.ForeignKey(
                        on_delete=models.PROTECT,
                        related_name="child_items",
                        to="db.collection",
                    ),
                ),
                ("parent_category", models.CharField(max_length=255)),
                ("data", models.JSONField(blank=True, default=dict)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("removed_at", models.DateTimeField(blank=True, null=True)),
                (
                    "artifact",
                    models.ForeignKey(
                        null=True,
                        on_delete=models.PROTECT,
                        related_name="collection_items",
                        to="db.artifact",
                    ),
                ),
                (
                    "collection",
                    models.ForeignKey(
                        null=True,
                        on_delete=models.PROTECT,
                        related_name="collection_items",
                        to="db.collection",
                    ),
                ),
                (
                    "created_by_user",
                    models.ForeignKey(
                        on_delete=models.PROTECT,
                        related_name="user_created_%(class)s",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "removed_by_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=models.PROTECT,
                        related_name="user_removed_%(class)s",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "constraints": [
                    models.CheckConstraint(
                        check=models.Q(
                            ("collection", models.F("parent_collection")),
                            _negated=True,
                        ),
                        name="db_collectionitem_distinct_parent_collection",
                    ),
                    models.CheckConstraint(
                        check=models.Q(
                            models.Q(
                                ("artifact__isnull", True),
                                ("child_type", "b"),
                                ("collection__isnull", True),
                            ),
                            models.Q(
                                ("child_type", "a"),
                                ("collection__isnull", True),
                                models.Q(
                                    ("artifact__isnull", False),
                                    ("removed_at__isnull", False),
                                    _connector="OR",
                                ),
                            ),
                            models.Q(
                                ("artifact__isnull", True),
                                ("child_type", "c"),
                                models.Q(
                                    ("collection__isnull", False),
                                    ("removed_at__isnull", False),
                                    _connector="OR",
                                ),
                            ),
                            _connector="OR",
                        ),
                        name="db_collectionitem_childtype_removedat_consistent",
                    ),
                ],
                "indexes": [
                    models.Index(
                        models.F("parent_collection"),
                        KeyTextTransform("package", "data"),
                        KeyTextTransform("version", "data"),
                        condition=models.Q(
                            ("category", ArtifactCategory["SOURCE_PACKAGE"]),
                            ("child_type", "a"),
                            ("parent_category", CollectionCategory["SUITE"]),
                        ),
                        name="db_ci_suite_source_idx",
                    ),
                    models.Index(
                        models.F("parent_collection"),
                        KeyTextTransform("srcpkg_name", "data"),
                        KeyTextTransform("srcpkg_version", "data"),
                        condition=models.Q(
                            ("category", ArtifactCategory["BINARY_PACKAGE"]),
                            ("child_type", "a"),
                            ("parent_category", CollectionCategory["SUITE"]),
                        ),
                        name="db_ci_suite_binary_source_idx",
                    ),
                ],
            },
        ),
        migrations.CreateModel(
            name="CollectionItemMatchConstraint",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "collection",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="item_match_constraints",
                        to="db.collection",
                    ),
                ),
                ("collection_item_id", models.BigIntegerField()),
                ("constraint_name", models.CharField(max_length=255)),
                ("key", models.TextField()),
                ("value", models.TextField()),
            ],
            options={
                "constraints": [
                    ExclusionConstraint(
                        expressions=(
                            (models.F("collection"), "="),
                            (models.F("constraint_name"), "="),
                            (models.F("key"), "="),
                            (models.F("value"), "<>"),
                        ),
                        name="db_collectionitemmatchconstraint_match_value",
                    )
                ],
                "indexes": [
                    models.Index(
                        fields=["collection_item_id"],
                        name="db_cimc_collection_item_idx",
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="FileInArtifact",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "artifact",
                    models.ForeignKey(
                        on_delete=models.PROTECT, to="db.artifact"
                    ),
                ),
                ("path", models.CharField(max_length=500)),
                (
                    "file",
                    models.ForeignKey(on_delete=models.PROTECT, to="db.file"),
                ),
                ("complete", models.BooleanField(default=False)),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(
                        fields=("artifact", "path"),
                        name="db_fileinartifact_unique_artifact_path",
                    )
                ]
            },
        ),
        migrations.CreateModel(
            name="FileInStore",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "store",
                    models.ForeignKey(
                        on_delete=models.PROTECT, to="db.filestore"
                    ),
                ),
                (
                    "file",
                    models.ForeignKey(on_delete=models.PROTECT, to="db.file"),
                ),
                ("data", models.JSONField(blank=True, default=dict)),
            ],
            options={
                "triggers": [
                    Trigger(
                        name="db_fileinstore_total_size_insert",
                        sql=UpsertTriggerSql(
                            func="UPDATE db_filestore SET total_size = total_size + db_file.size FROM db_file WHERE db_filestore.id = NEW.store_id AND db_file.id = NEW.file_id; RETURN NULL;",
                            hash="93dd6bf857752d47e15ce24887b0151f07f5aa77",
                            operation="INSERT",
                            pgid="pgtrigger_db_fileinstore_total_size_insert_8cd7f",
                            table="db_fileinstore",
                            when="AFTER",
                        ),
                    ),
                    Trigger(
                        name="db_fileinstore_total_size_delete",
                        sql=UpsertTriggerSql(
                            func="UPDATE db_filestore SET total_size = total_size - db_file.size FROM db_file WHERE db_filestore.id = OLD.store_id AND db_file.id = OLD.file_id; RETURN NULL;",
                            hash="1d8bad11847895fe36cd12b11d4f1a10fd451da7",
                            operation="DELETE",
                            pgid="pgtrigger_db_fileinstore_total_size_delete_07c98",
                            table="db_fileinstore",
                            when="AFTER",
                        ),
                    ),
                ],
                "constraints": [
                    models.UniqueConstraint(
                        fields=("store", "file"),
                        name="db_fileinstore_unique_store_file",
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="FileUpload",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "path",
                    models.CharField(
                        help_text="Path in the uploads directory",
                        max_length=500,
                        unique=True,
                    ),
                ),
                ("last_activity_at", models.DateTimeField(auto_now_add=True)),
                (
                    "file_in_artifact",
                    models.OneToOneField(
                        on_delete=models.PROTECT, to="db.fileinartifact"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Identity",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=models.SET_NULL,
                        related_name="identities",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "issuer",
                    models.CharField(
                        help_text="identifier of auhoritative system for this identity",
                        max_length=512,
                    ),
                ),
                (
                    "subject",
                    models.CharField(
                        help_text="identifier of the user in the issuer system",
                        max_length=512,
                    ),
                ),
                (
                    "last_used",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="last time this identity has been used",
                    ),
                ),
                ("claims", models.JSONField(default=dict)),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(
                        fields=("issuer", "subject"),
                        name="db_identity_unique_issuer_subject",
                    )
                ]
            },
        ),
        migrations.CreateModel(
            name="NotificationChannel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        max_length=20,
                        unique=True,
                        validators=[validate_notification_channel_name],
                    ),
                ),
                (
                    "method",
                    models.CharField(
                        choices=[("email", "Email")], max_length=10
                    ),
                ),
                ("data", models.JSONField(blank=True, default=dict)),
            ],
        ),
        migrations.CreateModel(
            name="Token",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "hash",
                    models.CharField(
                        max_length=64,
                        unique=True,
                        validators=[
                            MaxLengthValidator(64),
                            MinLengthValidator(64),
                        ],
                        verbose_name="Hexadecimal hash, length is 64 chars",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=models.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "comment",
                    models.CharField(
                        blank=True,
                        default="",
                        max_length=100,
                        verbose_name="Reason that this token was created",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("expire_at", models.DateTimeField(blank=True, null=True)),
                ("enabled", models.BooleanField(default=False)),
                (
                    "last_seen_at",
                    models.DateTimeField(
                        blank=True,
                        help_text="Last time that the token was used",
                        null=True,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="WorkerPool",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.SlugField(
                        help_text="Human readable name of the worker pool",
                        unique=True,
                    ),
                ),
                ("enabled", models.BooleanField(default=True)),
                ("architectures", models.JSONField(default=list)),
                ("tags", models.JSONField(blank=True, default=list)),
                ("specifications", models.JSONField(default=dict)),
                ("instance_wide", models.BooleanField(default=True)),
                ("ephemeral", models.BooleanField(default=False)),
                ("limits", models.JSONField(blank=True, default=dict)),
                ("registered_at", models.DateTimeField()),
                (
                    "provider_account",
                    models.ForeignKey(on_delete=models.PROTECT, to="db.asset"),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ScopeWorkerPool",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("priority", models.IntegerField(default=0)),
                ("limits", models.JSONField(blank=True, default=dict)),
                (
                    "scope",
                    models.ForeignKey(on_delete=models.CASCADE, to="db.scope"),
                ),
                (
                    "worker_pool",
                    models.ForeignKey(
                        on_delete=models.CASCADE, to="db.workerpool"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Worker",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.SlugField(
                        help_text="Human readable name of the worker based on the FQDN",
                        unique=True,
                    ),
                ),
                ("registered_at", models.DateTimeField()),
                ("connected_at", models.DateTimeField(blank=True, null=True)),
                (
                    "instance_created_at",
                    models.DateTimeField(blank=True, null=True),
                ),
                (
                    "token",
                    models.OneToOneField(
                        null=True,
                        on_delete=models.PROTECT,
                        related_name="worker",
                        to="db.token",
                    ),
                ),
                (
                    "activation_token",
                    models.OneToOneField(
                        null=True,
                        on_delete=models.SET_NULL,
                        related_name="activating_worker",
                        to="db.token",
                    ),
                ),
                ("static_metadata", models.JSONField(blank=True, default=dict)),
                (
                    "dynamic_metadata",
                    models.JSONField(blank=True, default=dict),
                ),
                (
                    "dynamic_metadata_updated_at",
                    models.DateTimeField(blank=True, null=True),
                ),
                (
                    "worker_type",
                    models.CharField(
                        choices=[
                            (WorkerType["EXTERNAL"], WorkerType["EXTERNAL"]),
                            (WorkerType["CELERY"], WorkerType["CELERY"]),
                            (WorkerType["SIGNING"], WorkerType["SIGNING"]),
                        ],
                        default=WorkerType["EXTERNAL"],
                        editable=False,
                        max_length=8,
                    ),
                ),
                (
                    "concurrency",
                    models.PositiveIntegerField(
                        default=1,
                        help_text="Number of tasks this worker can run simultaneously",
                    ),
                ),
                (
                    "worker_pool",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=models.PROTECT,
                        to="db.workerpool",
                    ),
                ),
                ("worker_pool_data", models.JSONField(blank=True, null=True)),
            ],
            options={
                "constraints": [
                    models.CheckConstraint(
                        check=models.Q(
                            ("worker_type", WorkerType["CELERY"]),
                            ("activation_token__isnull", False),
                            ("token__isnull", False),
                            _connector="OR",
                        ),
                        name="db_worker_celery_or_token",
                    )
                ]
            },
        ),
        migrations.CreateModel(
            name="WorkerPoolStatistics",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                ("runtime", models.IntegerField()),
                (
                    "worker",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=models.SET_NULL,
                        to="db.worker",
                    ),
                ),
                (
                    "worker_pool",
                    models.ForeignKey(
                        on_delete=models.CASCADE, to="db.workerpool"
                    ),
                ),
            ],
            options={
                "indexes": [
                    models.Index(
                        models.F("timestamp"), name="db_worker_pool_stat_ts_idx"
                    )
                ]
            },
        ),
        migrations.CreateModel(
            name="WorkerPoolTaskExecutionStatistics",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                ("runtime", models.IntegerField()),
                (
                    "scope",
                    models.ForeignKey(on_delete=models.CASCADE, to="db.scope"),
                ),
                (
                    "worker",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=models.SET_NULL,
                        to="db.worker",
                    ),
                ),
                (
                    "worker_pool",
                    models.ForeignKey(
                        on_delete=models.CASCADE, to="db.workerpool"
                    ),
                ),
            ],
            options={
                "indexes": [
                    models.Index(
                        models.F("timestamp"), name="db_worker_pool_exec_ts_idx"
                    )
                ]
            },
        ),
        migrations.CreateModel(
            name="WorkflowTemplate",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                (
                    "workspace",
                    models.ForeignKey(
                        on_delete=models.PROTECT, to="db.workspace"
                    ),
                ),
                (
                    "task_name",
                    models.CharField(
                        max_length=100,
                        verbose_name="Name of the Workflow orchestrator class",
                    ),
                ),
                (
                    "task_data",
                    models.JSONField(
                        blank=True, default=dict, encoder=DjangoJSONEncoder
                    ),
                ),
                (
                    "priority",
                    models.IntegerField(
                        default=0,
                        help_text="Base priority for work requests created from this template",
                    ),
                ),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(
                        fields=("name", "workspace"),
                        name="db_workflowtemplate_unique_name_workspace",
                    )
                ]
            },
        ),
        migrations.CreateModel(
            name="WorkRequest",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "workspace",
                    models.ForeignKey(
                        on_delete=models.PROTECT, to="db.workspace"
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("started_at", models.DateTimeField(blank=True, null=True)),
                ("completed_at", models.DateTimeField(blank=True, null=True)),
                (
                    "workflow_last_activity_at",
                    models.DateTimeField(blank=True, null=True),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=models.PROTECT, to=settings.AUTH_USER_MODEL
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("pending", "Pending"),
                            ("running", "Running"),
                            ("completed", "Completed"),
                            ("aborted", "Aborted"),
                            ("blocked", "Blocked"),
                        ],
                        default="pending",
                        editable=False,
                        max_length=9,
                    ),
                ),
                (
                    "workflow_runtime_status",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("needs_input", "Needs Input"),
                            ("running", "Running"),
                            ("waiting", "Waiting"),
                            ("pending", "Pending"),
                            ("aborted", "Aborted"),
                            ("completed", "Completed"),
                            ("blocked", "Blocked"),
                        ],
                        editable=False,
                        max_length=11,
                        null=True,
                    ),
                ),
                ("workflows_need_update", models.BooleanField(default=False)),
                (
                    "result",
                    models.CharField(
                        choices=[
                            ("", ""),
                            ("success", "Success"),
                            ("failure", "Failure"),
                            ("error", "Error"),
                        ],
                        default="",
                        editable=False,
                        max_length=7,
                    ),
                ),
                (
                    "worker",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=models.CASCADE,
                        related_name="assigned_work_requests",
                        to="db.worker",
                    ),
                ),
                (
                    "task_type",
                    models.CharField(
                        choices=[
                            (TaskTypes["WORKER"], TaskTypes["WORKER"]),
                            (TaskTypes["SERVER"], TaskTypes["SERVER"]),
                            (TaskTypes["INTERNAL"], TaskTypes["INTERNAL"]),
                            (TaskTypes["WORKFLOW"], TaskTypes["WORKFLOW"]),
                            (TaskTypes["SIGNING"], TaskTypes["SIGNING"]),
                            (TaskTypes["WAIT"], TaskTypes["WAIT"]),
                        ],
                        default=TaskTypes["WORKER"],
                        editable=False,
                        max_length=8,
                        verbose_name="Type of task to execute",
                    ),
                ),
                (
                    "task_name",
                    models.CharField(
                        max_length=100,
                        verbose_name="Name of the task to execute",
                    ),
                ),
                (
                    "task_data",
                    models.JSONField(
                        blank=True, default=dict, encoder=DjangoJSONEncoder
                    ),
                ),
                (
                    "dynamic_task_data",
                    models.JSONField(
                        blank=True, encoder=DjangoJSONEncoder, null=True
                    ),
                ),
                (
                    "priority_base",
                    models.IntegerField(
                        default=0,
                        help_text="Base priority of this work request",
                    ),
                ),
                (
                    "priority_adjustment",
                    models.IntegerField(
                        default=0,
                        help_text="Administrator adjustment to the priority of this work request",
                    ),
                ),
                (
                    "output_data_json",
                    models.JSONField(
                        blank=True,
                        db_column="output_data",
                        encoder=DjangoJSONEncoder,
                        null=True,
                    ),
                ),
                (
                    "unblock_strategy",
                    models.CharField(
                        choices=[
                            ("deps", "Dependencies have completed"),
                            ("manual", "Manually unblocked"),
                        ],
                        default="deps",
                        editable=False,
                        max_length=6,
                    ),
                ),
                (
                    "workflow_template",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=models.SET_NULL,
                        to="db.workflowtemplate",
                    ),
                ),
                (
                    "workflow_data_json",
                    models.JSONField(
                        blank=True,
                        db_column="workflow_data",
                        default=dict,
                        encoder=DjangoJSONEncoder,
                    ),
                ),
                (
                    "event_reactions_json",
                    models.JSONField(
                        blank=True,
                        db_column="event_reactions",
                        default=dict,
                        encoder=DjangoJSONEncoder,
                    ),
                ),
                (
                    "internal_collection",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=models.PROTECT,
                        related_name="workflow",
                        to="db.collection",
                    ),
                ),
                (
                    "expiration_delay",
                    models.DurationField(blank=True, null=True),
                ),
                (
                    "supersedes",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=models.SET_NULL,
                        related_name="superseded",
                        to="db.workrequest",
                    ),
                ),
                (
                    "configured_task_data",
                    models.JSONField(
                        blank=True,
                        encoder=DjangoJSONEncoder,
                        help_text="task_data with task configuration applied",
                        null=True,
                    ),
                ),
                (
                    "version",
                    models.IntegerField(
                        default=0,
                        help_text="version of the code that computed the dynamic task data",
                    ),
                ),
            ],
            options={
                "indexes": [
                    models.Index(
                        models.OrderBy(
                            CombinedExpression(
                                models.F("priority_base"),
                                "+",
                                models.F("priority_adjustment"),
                            ),
                            descending=True,
                        ),
                        models.F("created_at"),
                        condition=models.Q(("status", "pending")),
                        name="db_workrequest_pending_idx",
                    ),
                    models.Index(
                        models.F("worker"),
                        condition=models.Q(
                            ("status__in", ("pending", "running"))
                        ),
                        name="db_workrequest_worker_idx",
                    ),
                    models.Index(
                        models.F("workflows_need_update"),
                        name="db_workrequest_wf_update_idx",
                    ),
                ],
                "permissions": [
                    (
                        "manage_workrequest_priorities",
                        "Can set positive priority adjustments on work requests",
                    )
                ],
                "ordering": ["id"],
            },
        ),
        migrations.AddField(
            model_name="asset",
            name="created_by_work_request",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=models.SET_NULL,
                to="db.workrequest",
            ),
        ),
        migrations.AddConstraint(
            model_name="asset",
            constraint=JsonDataUniqueConstraint(
                condition=models.Q(
                    ("category", AssetCategory["CLOUD_PROVIDER_ACCOUNT"])
                ),
                fields=("data->>'name'",),
                name="db_asset_unique_cloud_provider_acct_name",
                nulls_distinct=False,
            ),
        ),
        migrations.AddConstraint(
            model_name="asset",
            constraint=JsonDataUniqueConstraint(
                condition=models.Q(("category", AssetCategory["SIGNING_KEY"])),
                fields=("data->>'fingerprint'",),
                name="db_asset_unique_signing_key_fingerprints",
                nulls_distinct=False,
            ),
        ),
        migrations.AddField(
            model_name="artifact",
            name="created_by_work_request",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=models.SET_NULL,
                to="db.workrequest",
            ),
        ),
        migrations.AddField(
            model_name="artifact",
            name="files",
            field=models.ManyToManyField(
                through="db.FileInArtifact", to="db.file"
            ),
        ),
        migrations.AddField(
            model_name="collection",
            name="child_artifacts",
            field=models.ManyToManyField(
                related_name="parent_collections",
                through="db.CollectionItem",
                to="db.artifact",
            ),
        ),
        migrations.AddField(
            model_name="collection",
            name="child_collections",
            field=models.ManyToManyField(
                related_name="parent_collections",
                through="db.CollectionItem",
                to="db.collection",
            ),
        ),
        migrations.AddField(
            model_name="collectionitem",
            name="created_by_workflow",
            field=models.ForeignKey(
                null=True,
                on_delete=models.PROTECT,
                related_name="workflow_created_%(class)s",
                to="db.workrequest",
            ),
        ),
        migrations.AddField(
            model_name="collectionitem",
            name="removed_by_workflow",
            field=models.ForeignKey(
                null=True,
                on_delete=models.PROTECT,
                related_name="workflow_removed_%(class)s",
                to="db.workrequest",
            ),
        ),
        migrations.AddConstraint(
            model_name="collectionitem",
            constraint=models.UniqueConstraint(
                condition=models.Q(("removed_at__isnull", True)),
                fields=("name", "parent_collection"),
                name="db_collectionitem_unique_active_name",
            ),
        ),
        migrations.AddConstraint(
            model_name="collectionitem",
            constraint=JsonDataUniqueConstraint(
                condition=models.Q(
                    ("parent_category", CollectionCategory["ENVIRONMENTS"]),
                    ("removed_at__isnull", True),
                ),
                fields=(
                    "category",
                    "data->>'codename'",
                    "data->>'architecture'",
                    "data->>'variant'",
                    "data->>'backend'",
                    "parent_collection",
                ),
                name="db_collectionitem_unique_debian_environments",
                nulls_distinct=False,
            ),
        ),
        migrations.AddField(
            model_name="filestore",
            name="files",
            field=models.ManyToManyField(
                through="db.FileInStore", to="db.file"
            ),
        ),
        migrations.AddField(
            model_name="filestore",
            name="provider_account",
            field=models.ForeignKey(
                blank=True, null=True, on_delete=models.PROTECT, to="db.asset"
            ),
        ),
        migrations.AddField(
            model_name="group",
            name="users",
            field=models.ManyToManyField(
                blank=True,
                related_name="debusine_groups",
                through="db.GroupMembership",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="scope",
            name="file_stores",
            field=models.ManyToManyField(
                related_name="scopes",
                through="db.FileStoreInScope",
                to="db.filestore",
            ),
        ),
        migrations.AddField(
            model_name="workrequest",
            name="dependencies",
            field=models.ManyToManyField(
                related_name="reverse_dependencies", to="db.workrequest"
            ),
        ),
        migrations.AddField(
            model_name="workrequest",
            name="parent",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=models.CASCADE,
                related_name="children",
                to="db.workrequest",
            ),
        ),
        migrations.AddField(
            model_name="workspace",
            name="inherits",
            field=models.ManyToManyField(
                related_name="inherited_by",
                through="db.WorkspaceChain",
                to="db.workspace",
            ),
        ),
        migrations.RunPython(create_default_objects),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the debusine-client interactive data entry code."""

import re
import textwrap
from typing import Any, override
from unittest import mock

from debusine.client.dataentry import DataEntry
from debusine.client.tests.utils import BufferConsole
from debusine.test import TestCase


class MockDataEntry(DataEntry):
    """Concrete DataEntry implementation tweaked for tests."""

    console: BufferConsole

    def __init__(self) -> None:
        """Set up the rich console to output plaing text to self.output."""
        super().__init__(console=BufferConsole())

    @override
    def menu(self) -> None:
        """Mock menu implementation."""
        self.console.print("__menu__")


class DataEntryTests(TestCase):
    """Tests for :py:class:`DataEntry`."""

    DEFAULT_PROMPT: str = "__prompt__"

    def assertInputLine(
        self,
        de: MockDataEntry,
        input_lines: list[str],
        output_lines: list[str],
        result: str | None,
        prompt: str = DEFAULT_PROMPT,
        **kwargs: Any,
    ) -> None:
        """Run an input_line session on the DataEntry object."""
        de.console.reset_output()
        with mock.patch(
            "debusine.client.dataentry.input", side_effect=iter(input_lines)
        ):
            res = de.input_line(prompt, **kwargs)
        self.assertEqual(res, result)
        self.assertEqual(
            de.console.output.getvalue().splitlines(), output_lines
        )

    def test_input_line_plain(self) -> None:
        """Test plain input_line."""
        de = MockDataEntry()
        with mock.patch(
            "debusine.client.dataentry.readline.set_startup_hook"
        ) as readline_startup_hook:
            self.assertInputLine(de, ["test"], [self.DEFAULT_PROMPT], "test")
        self.assertEqual(len(readline_startup_hook.call_args_list), 1)

    def test_input_line_empty(self) -> None:
        """Test input_line with an empty value."""
        de = MockDataEntry()
        self.assertInputLine(de, [""], [self.DEFAULT_PROMPT], None)

    def test_input_line_required(self) -> None:
        """Test input_line requiring a value."""
        de = MockDataEntry()
        self.assertInputLine(
            de,
            ["", "test"],
            [
                self.DEFAULT_PROMPT,
                "Please enter a value: this field is needed",
                self.DEFAULT_PROMPT,
            ],
            "test",
            required=True,
        )

    def test_input_line_validator(self) -> None:
        """Test input_line with validation."""
        de = MockDataEntry()
        self.assertInputLine(
            de,
            ["1234", "test"],
            [
                self.DEFAULT_PROMPT,
                "'1234' should match ^[a-z]+$",
                self.DEFAULT_PROMPT,
            ],
            "test",
            validator=re.compile(r"^[a-z]+$"),
        )

    def test_input_line_initial(self) -> None:
        """Test input_line initial value."""
        de = MockDataEntry()
        with mock.patch(
            "debusine.client.dataentry.readline.set_startup_hook"
        ) as readline_startup_hook:
            self.assertInputLine(
                de, ["test"], [self.DEFAULT_PROMPT], "test", initial="foo"
            )

        self.assertEqual(len(readline_startup_hook.call_args_list), 2)

    def test_help(self) -> None:
        """Test help generation."""
        de = MockDataEntry()
        self.assertFalse(de.onecmd("help"))
        self.assertEqual(
            de.console.output.getvalue(),
            textwrap.dedent(
                """
            Command summary (using the first letter is enough):

             * eof: Stop processing.
             * help: Command line help.
             * quit: Quit editor.

            You can use 'help' command for detailed help
               """
            ),
        )

    def test_help_missing(self) -> None:
        """Test help for a missing command."""
        de = MockDataEntry()
        self.assertFalse(de.onecmd("help missing"))
        self.assertEqual(
            de.console.output.getvalue(),
            "\nNo help found for 'missing'\n",
        )

    def test_help_command(self) -> None:
        """Test help for an existing command."""
        de = MockDataEntry()
        self.assertFalse(de.onecmd("help quit"))
        self.assertEqual(
            de.console.output.getvalue(),
            "\nQuit editor.\n\n",
        )

    def test_quit(self) -> None:
        """Quit signals to end the Cmd loop."""
        de = MockDataEntry()
        self.assertTrue(de.onecmd("quit"))
        self.assertEqual(de.console.output.getvalue(), "")

    def test_EOF(self) -> None:
        """EOF signals to end the Cmd loop."""
        de = MockDataEntry()
        with mock.patch.object(de, "do_quit", return_value=True) as do_quit:
            self.assertTrue(de.do_eof(""))
        do_quit.assert_called()
        self.assertEqual(de.console.output.getvalue(), "\n")

    def test_default(self) -> None:
        """Test handling unrecognized commands."""
        de = MockDataEntry()
        self.assertFalse(de.onecmd("mischief"))
        self.assertEqual(
            de.console.output.getvalue(), "command not recognized: 'mischief'\n"
        )

    def test_auto_menu(self) -> None:
        """Test printing menu at the end of each command."""
        de = MockDataEntry()
        self.assertFalse(de.postcmd(False, ""))
        self.assertEqual(de.console.output.getvalue(), "__menu__\n")
        de.console.reset_output()
        self.assertTrue(de.postcmd(True, ""))
        self.assertEqual(de.console.output.getvalue(), "")

    def test_cmdloop_menu(self) -> None:
        """Test printing menu at the beginning of cmdloop."""
        de = MockDataEntry()
        with (
            mock.patch.object(de, "menu") as menu,
            mock.patch("debusine.client.dataentry.cmd.Cmd.cmdloop") as cmdloop,
        ):
            de.cmdloop()

        menu.assert_called()
        cmdloop.assert_called()

    def test_abbreviations(self) -> None:
        """Test abbreviating commands."""
        de = MockDataEntry()
        for entered, expected in (
            ("", ""),
            ("h", "help"),
            ("he", "help"),
            ("hel", "help"),  # codespell:ignore hel
            ("help", "help"),
            ("h foo", "help foo"),
            ("q", "quit"),
            ("q foo", "quit foo"),
            ("x", "x"),
            ("x foo", "x foo"),
        ):
            with self.subTest(entered=entered):
                self.assertEqual(de.precmd(entered), expected)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the debusine Cli WorkflowTemplate commands."""

from collections.abc import Callable
from functools import partial
from textwrap import dedent
from typing import Any, override
from unittest import mock

import yaml

from debusine.client.commands.base import Command, DebusineCommand
from debusine.client.commands.tests.base import BaseCliTests
from debusine.client.commands.workflow_templates import WorkflowTemplateCommand
from debusine.client.debusine import Debusine
from debusine.client.exceptions import DebusineError
from debusine.client.models import WorkflowTemplateData, WorkflowTemplateDataNew
from debusine.utils.input import YamlEditor


class WorkflowTemplateCommandTests(BaseCliTests):
    """Tests for the :py:class:`CollectionCommand` class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.data = WorkflowTemplateData(
            id=42,
            name="sample",
            task_name="noop",
            static_parameters={"test": 2},
            runtime_parameters={},
            priority=42,
            links={"webui_self": "/wt/42"},
        )

    @Command.preserve_registry()
    def _command(self, **kwargs: Any) -> WorkflowTemplateCommand:
        class ConcreteWorkflowTemplateCommand(WorkflowTemplateCommand):
            """Version of WorkflowTemplateCommand that can be instantiated."""

            @override
            def run(self) -> None:
                raise NotImplementedError()

        return self.command(
            ConcreteWorkflowTemplateCommand, workspace="workspace"
        )

    def test_list_rich(self) -> None:
        command = self._command(workspace="workspace")
        with (
            mock.patch(
                "debusine.client.commands.workflow_templates.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.workflow_templates.rich.print"
            ) as rich_print,
        ):
            stderr, stdout = self.capture_output(
                partial(command._list_rich, [self.data])
            )

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        rich_print.assert_called_once()

        self.assertEqual(table.add_row.call_count, 1)
        url = "https://debusine.debian.org/wt/42"
        call = table.add_row.call_args_list[0]
        self.assertEqual(len(call.args), 8)
        self.assertEqual(call.args[0], "42")
        self.assertEqual(call.args[1], f"[link={url}]sample[/]")
        self.assertEqual(call.args[2], "noop")
        self.assertEqual(call.args[3], "42")
        self.assertEqual(call.args[7], "False")

    def test_show_rich(self) -> None:
        command = self._command(workspace="workspace")
        with (
            mock.patch(
                "debusine.client.commands.workflow_templates.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.workflow_templates.rich.print"
            ) as rich_print,
        ):
            stderr, stdout = self.capture_output(
                partial(command._show_rich, self.data)
            )

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        rich_print.assert_called_once()

        self.assertEqual(table.add_row.call_count, 8)
        url = "https://debusine.debian.org/wt/42"
        rows = [
            call_args_list.args
            for call_args_list in table.add_row.call_args_list
        ]
        self.assertEqual(rows[0], ("ID:", "#42"))
        self.assertEqual(rows[1], ("Name:", f"[link={url}]sample[/]"))
        self.assertEqual(rows[2], ("Task name:", "noop"))
        self.assertEqual(rows[3], ("Priority:", "42"))
        self.assertEqual(rows[6], ("Extra groups:", ""))
        self.assertEqual(rows[7], ("Restricted:", "False"))

    def test_parse_input_data_validation(self) -> None:
        defaults = {
            "static_parameters": {},
            "runtime_parameters": "any",
            "extra_groups": [],
            "restricted": True,
        }
        for data, error in (
            (
                {**defaults, "static_parameters": 42},
                "Error: static_parameters must be a dictionary. It is: int",
            ),
            (
                {**defaults, "runtime_parameters": 42},
                "Error: runtime_parameters must be a dictionary or string."
                " It is: int",
            ),
            (
                {**defaults, "extra_groups": 42},
                "Error: extra_groups must be a list. It is: int",
            ),
            (
                {**defaults, "extra_groups": ["foo", []]},
                "Error: extra_groups must be a list of integers or strings",
            ),
            (
                {**defaults, "restricted": 42},
                "Error: restricted must be a boolean. It is: int",
            ),
        ):
            with (
                self.subTest(data=repr(data)),
                mock.patch(
                    "debusine.client.commands.workflow_templates"
                    ".WorkflowTemplateCommand._fail",
                    side_effect=RuntimeError("_fail"),
                ) as fail,
            ):
                with self.assertRaisesRegex(RuntimeError, "_fail"):
                    WorkflowTemplateCommand._parse_input_data(data)
                fail.assert_called_once_with(error)


class CreateTests(BaseCliTests):
    """Tests for the CLI :py:class:`workflow_template.Create` class."""

    def _command(
        self, task_name: str, template_name: str, *args: str
    ) -> list[str]:
        """Build a command line."""
        return ["workflow-template", "create", task_name, template_name, *args]

    def assert_create_success(
        self,
        task_name: str,
        template_name: str,
        *args: str,
        workspace: str,
        data: WorkflowTemplateDataNew,
    ) -> None:
        """
        Call workflow-template create with the given arguments.

        :param args: arguments to append to
          ``debusine workflow-template create``
        :param data: data expected to be sent to the server
        """
        command = self.create_command(
            self._command(task_name, template_name, "--yaml", *args)
        )
        assert isinstance(command, DebusineCommand)
        result = WorkflowTemplateData(id=42, **data.model_dump())
        result.links = {"self": "/api/wt/42", "webui_self": "/wt/42"}
        with mock.patch.object(
            command.debusine, "workflow_template_create", return_value=result
        ) as create:
            stderr, stdout = self.capture_output(command.run)

        create.assert_called_with(workspace, data)
        self.assertEqual(stderr, "")
        self.assertStdoutYaml(workspace, stdout, result)

    def assertStdoutYaml(
        self,
        workspace: str,  # noqa: ARG002, U100
        stdout: str,
        result: WorkflowTemplateData,
    ) -> None:
        """Ensure the program stdout matches the given result."""
        actual = WorkflowTemplateData.model_validate(yaml.safe_load(stdout))
        self.assertEqual(actual, result)

    def test_invalid_task_name(self) -> None:
        """CLI fails if the task name is bad."""
        task_name = "task-name"

        self.enterContext(self.patch_sys_stdin_read("{}"))
        cli = self.create_cli(self._command(task_name, "test"))

        with mock.patch.object(
            Debusine,
            "workflow_template_create",
            autospec=True,
            side_effect=DebusineError(title=f"invalid {task_name}"),
        ):
            exception = self.assertShowsError(cli.execute)

        self.assertDebusineError(exception, {"title": f"invalid {task_name}"})

    def test_specific_workspace(self) -> None:
        """CLI parses the command line and uses --workspace."""
        self.enterContext(self.patch_sys_stdin_read("{}"))
        self.assert_create_success(
            "sbuild",
            "sbuild-test",
            "--workspace",
            "Testing",
            workspace="Testing",
            data=WorkflowTemplateDataNew(
                name="sbuild-test", task_name="sbuild", static_parameters={}
            ),
        )

    def test_success_data_from_file(self) -> None:
        """CLI creates a workflow template with data from a file."""
        tempfile = self.create_temporary_file()
        tempfile.write_text("{}")
        self.assert_create_success(
            "sbuild",
            "sbuild-test",
            f"--data={tempfile}",
            workspace="developers",
            data=WorkflowTemplateDataNew(
                name="sbuild-test",
                task_name="sbuild",
                static_parameters={},
                runtime_parameters={},
            ),
        )

    def test_success_data_from_stdin(self) -> None:
        """CLI creates a workflow template with data from stdin."""
        self.enterContext(self.patch_sys_stdin_read("{}"))
        self.assert_create_success(
            "sbuild",
            "sbuild-test",
            workspace="developers",
            data=WorkflowTemplateDataNew(
                name="sbuild-test",
                task_name="sbuild",
                static_parameters={},
                runtime_parameters={},
            ),
        )

    def test_parameters_are_parsed(self) -> None:
        """CLI creates a workflow template with specified parameters."""
        tempfile = self.create_temporary_file()
        expected = WorkflowTemplateDataNew(
            name="sbuild-test",
            task_name="sbuild",
            static_parameters={"foo": "bar"},
            runtime_parameters={"foo": ["bar", "baz"]},
        )
        data = {
            "static_parameters": expected.static_parameters,
            "runtime_parameters": expected.runtime_parameters,
        }
        tempfile.write_text(yaml.safe_dump(data))
        self.assert_create_success(
            "sbuild",
            "sbuild-test",
            f"--data={tempfile}",
            workspace="developers",
            data=expected,
        )

    def test_data_is_empty(self) -> None:
        """CLI rejects a workflow template with empty data."""
        empty_file = self.create_temporary_file()
        cli = self.create_cli(
            self._command("sbuild", "sbuild-test", "--data", str(empty_file))
        )

        stderr, stdout = self.capture_output(
            cli.execute, assert_system_exit_code=3
        )

        self.assertEqual(
            stderr, "Error: data must be a dictionary. It is empty\n"
        )
        self.assertEqual(stdout, "")

    def test_yaml_errors_failed(self) -> None:
        """cli.execute() deals with different invalid data."""
        workflow_templates = [
            {
                "data": dedent(
                    """\
                    test:
                      name: a-name
                        first-name: some first name
                    """
                ),
                "comment": "yaml.safe_load raises ScannerError",
            },
            {
                "data": dedent(
                    """\
                    input:
                      source_url: https://example.com
                     )
                    """
                ),
                "comment": "yaml.safe_load raises ParserError",
            },
        ]

        for workflow_template in workflow_templates:
            data = workflow_template["data"]
            with (
                self.subTest(data),
                self.patch_sys_stdin_read(data),
            ):
                cli = self.create_cli(
                    self._command("task-name", "template-name")
                )
                stderr, _ = self.capture_output(
                    cli.execute, assert_system_exit_code=3
                )

                self.assertRegex(stderr, "^Error parsing YAML:")
                self.assertRegex(stderr, "Fix the YAML data\n$")


class ListTests(BaseCliTests):
    """Tests for the :py:class:`workflow_templates.List` class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.data = WorkflowTemplateData(
            id=42,
            name="sample",
            task_name="noop",
            static_parameters={"test": 2},
            runtime_parameters={},
            priority=42,
        )

    def test_list(self) -> None:
        command = self.create_command(
            ["workflow-template", "list", "--workspace=workspace"]
        )
        assert isinstance(command, DebusineCommand)
        with (
            mock.patch.object(
                command.debusine,
                "workflow_template_iter",
                return_value=[self.data],
            ) as collection_iter,
            mock.patch.object(command, "list") as list_,
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        collection_iter.assert_called_once_with("workspace")
        list_.assert_called_once_with([self.data])


class ShowTests(BaseCliTests):
    """Tests for the :py:class:`workflow_template.Show` class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.data = WorkflowTemplateData(
            id=42,
            name="sample",
            task_name="noop",
            static_parameters={"test": 2},
            runtime_parameters={},
            priority=42,
        )

    def test_show_id(self) -> None:
        command = self.create_command(
            ["workflow-template", "show", "--workspace=workspace", "42"]
        )
        assert isinstance(command, DebusineCommand)
        with (
            mock.patch.object(
                command.debusine,
                "workflow_template_get",
                return_value=self.data,
            ) as get,
            mock.patch.object(command, "show") as show,
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        get.assert_called_once_with("workspace", "42")
        show.assert_called_once_with(self.data)

    def test_show_name(self) -> None:
        command = self.create_command(
            ["workflow-template", "show", "--workspace=workspace", "sample"]
        )
        assert isinstance(command, DebusineCommand)
        with (
            mock.patch.object(
                command.debusine,
                "workflow_template_get",
                return_value=self.data,
            ) as get,
            mock.patch.object(command, "show") as show,
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        get.assert_called_once_with("workspace", "sample")
        show.assert_called_once_with(self.data)


class ManageTests(BaseCliTests):
    """Tests for the :py:class:`workflow_templates.Manage` class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.sample = WorkflowTemplateData(
            id=42,
            name="sample",
            task_name="noop",
            static_parameters={"test": 2},
            runtime_parameters={},
            priority=42,
        )
        self.edited = self.sample.model_copy()

    def call(
        self, *args: str, assert_system_exit_code: int | None = None
    ) -> tuple[mock.Mock, mock.Mock, str, str]:
        """Call workflow_template_update"""
        command = self.create_command(
            [
                "workflow-template",
                "manage",
                "--workspace=workspace",
                "sample",
                *args,
            ]
        )
        assert isinstance(command, DebusineCommand)

        with (
            mock.patch.object(
                command.debusine,
                "workflow_template_get",
                return_value=self.sample,
            ),
            mock.patch.object(
                command.debusine,
                "workflow_template_update",
                return_value=self.edited,
            ) as update,
            mock.patch.object(command, "show") as show,
        ):
            stderr, stdout = self.capture_output(
                command.run, assert_system_exit_code=assert_system_exit_code
            )
        return update, show, stdout, stderr

    def assertManages(self, *args: str) -> None:
        """Call workflow_template_update and match self.edited."""
        update, show, stdout, stderr = self.call(*args)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        update.assert_called_once_with("workspace", self.edited)
        show.assert_called_once_with(self.edited)

    def assertFails(self, *args: str, error: str) -> None:
        """Call workflow_template_update and match error to stderr."""
        update, show, stdout, stderr = self.call(
            *args, assert_system_exit_code=3
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, error)
        update.assert_not_called()
        show.assert_not_called()

    def test_rename(self) -> None:
        self.edited.name = "renamed"
        self.assertManages("--rename=renamed")

    def test_set_priority(self) -> None:
        self.edited.priority = 1
        self.assertManages("--priority=1")

    def test_set_data(self) -> None:
        infile = self.create_temporary_file()
        infile.write_text("{'static_parameters': {'test': 7}}")
        self.edited.static_parameters = {"test": 7}
        self.assertManages(f"--data={infile}")

    def test_set_data_stdin(self) -> None:
        self.enterContext(
            self.patch_sys_stdin_read("{'static_parameters': {'test': 7}}")
        )
        self.edited.static_parameters = {"test": 7}
        self.assertManages("--data=-")

    def test_set_data_unknown_key(self) -> None:
        infile = self.create_temporary_file()
        infile.write_text("{'unknown': 'parameter'}")
        self.assertFails(
            f"--data={infile}",
            error=(
                "Error parsing input data: Expecting only 4 keys: "
                "['extra_groups', 'restricted', 'runtime_parameters',"
                " 'static_parameters'],"
                " found: ['unknown']\n"
            ),
        )

    def test_set_data_static_parameters_garbage(self) -> None:
        infile = self.create_temporary_file()
        infile.write_text("{'static_parameters': 'garbage'}")
        self.assertFails(
            f"--data={infile}",
            error=(
                "Error: static_parameters must be a dictionary. It is: str\n"
            ),
        )

    def test_set_data_runtime_parameters_garbage(self) -> None:
        infile = self.create_temporary_file()
        infile.write_text("{'runtime_parameters': 123}")
        self.assertFails(
            f"--data={infile}",
            error=(
                "Error: runtime_parameters must be a dictionary or string. "
                "It is: int\n"
            ),
        )


class DeleteTests(BaseCliTests):
    """Tests for the :py:class:`workflow_templates.Delete` class."""

    def test_delete(self) -> None:
        """Call workflow_template_delete"""
        command = self.create_command(
            [
                "workflow-template",
                "delete",
                "--workspace=workspace",
                "sample",
            ]
        )
        assert isinstance(command, DebusineCommand)

        with (
            mock.patch.object(
                command.debusine,
                "workflow_template_delete",
            ) as workflow_template_delete,
            mock.patch.object(command, "feedback") as feedback,
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        workflow_template_delete.assert_called_once_with("workspace", "sample")
        feedback.assert_called_once_with("Workflow template deleted.")


class EditTests(BaseCliTests):
    """Tests for the :py:class:`workflow_templates.Edit` class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.sample = WorkflowTemplateData(
            id=42,
            name="sample",
            task_name="noop",
            static_parameters={"test": 2},
            runtime_parameters={},
            priority=42,
        )
        self.edited = self.sample.model_copy()

    def edit(
        self,
        *args: str,
        mock_edit: Callable[[YamlEditor[dict[str, Any]]], bool],
        assert_system_exit_code: int | None = None,
    ) -> tuple[mock.Mock, mock.Mock, str, str]:
        """Call workflow-template edit."""
        command = self.create_command(
            [
                "workflow-template",
                "edit",
                "--workspace=workspace",
                "sample",
                *args,
            ]
        )
        assert isinstance(command, DebusineCommand)

        with (
            mock.patch.object(
                command.debusine,
                "workflow_template_get",
                return_value=self.sample,
            ),
            mock.patch.object(
                command.debusine,
                "workflow_template_update",
                return_value=self.edited,
            ) as update,
            mock.patch.object(command, "show") as show,
            mock.patch(
                "debusine.utils.input.YamlEditor.edit",
                side_effect=mock_edit,
                autospec=True,
            ),
        ):
            stderr, stdout = self.capture_output(
                command.run, assert_system_exit_code=assert_system_exit_code
            )
        return (update, show, stdout, stderr)

    def assertEdits(
        self,
        *args: str,
        mock_edit: Callable[[YamlEditor[dict[str, Any]]], bool],
        unchanged: bool = False,
    ) -> None:
        """Call workflow-template edit and match self.edited."""
        update, show, stdout, stderr = self.edit(*args, mock_edit=mock_edit)

        self.assertEqual(stderr, "")
        if unchanged:
            update.assert_not_called()
            self.assertEqual(
                stdout, "Workflow template sample is left unchanged.\n"
            )
        else:
            update.assert_called_once_with("workspace", self.edited)
            self.assertEqual(stdout, "")
        show.assert_called_once_with(self.edited)

    def assertEditFails(
        self,
        *args: str,
        mock_edit: Callable[[YamlEditor[dict[str, Any]]], bool],
        error: str,
    ) -> None:
        """Call workflow-template edit and match error to stderr."""
        update, show, stdout, stderr = self.edit(
            *args, mock_edit=mock_edit, assert_system_exit_code=3
        )

        self.assertEqual(stderr, error)
        self.assertEqual(stdout, "")
        update.assert_not_called()
        show.assert_not_called()

    def test_edit_unchanged(self) -> None:
        def mock_edit(self_: YamlEditor[dict[str, Any]]) -> bool:
            self_.value = {
                "static_parameters": self.sample.static_parameters,
                "runtime_parameters": self.sample.runtime_parameters,
                "extra_groups": self.sample.extra_groups,
                "restricted": self.sample.restricted,
            }
            self_.cleanup()
            return True

        self.assertEdits(mock_edit=mock_edit, unchanged=True)

    def test_edit_changed(self) -> None:
        self.edited.static_parameters = {"edited": True}

        def mock_edit(self_: YamlEditor[dict[str, Any]]) -> bool:
            self_.value = {
                "static_parameters": self.edited.static_parameters,
                "runtime_parameters": self.edited.runtime_parameters,
                "extra_groups": self.edited.extra_groups,
                "restricted": self.edited.restricted,
            }
            self_.cleanup()
            return True

        self.assertEdits(mock_edit=mock_edit)

    def test_edit_aborted(self) -> None:
        def mock_edit(self_: YamlEditor[dict[str, Any]]) -> bool:
            self_.value = {"edited": True}
            self_.cleanup()
            return False

        self.assertEdits(mock_edit=mock_edit, unchanged=True)

    def test_edit_unexpected_keys(self) -> None:
        def mock_edit(self_: YamlEditor[dict[str, Any]]) -> bool:
            self_.value = {"garbage": ""}
            self_.cleanup()
            return True

        self.assertEditFails(
            mock_edit=mock_edit,
            error=(
                "Error parsing input data: Expecting 4 keys: "
                "['extra_groups', 'restricted', 'runtime_parameters',"
                " 'static_parameters'],"
                " found: ['garbage']\n"
            ),
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""The collection manager for debusine:task-configuration collections."""

import datetime as dt
from collections.abc import Generator, Iterable
from typing import Any, NamedTuple, override

from django.db import IntegrityError
from django.db.models import Case, JSONField, Q, Value, When
from django.db.models.functions import Coalesce
from django.utils import timezone

from debusine.artifacts.models import (
    BareDataCategory,
    CollectionCategory,
    DebusineTaskConfiguration,
)
from debusine.db.models import Collection, CollectionItem, User, WorkRequest
from debusine.db.orm_utils import CompareBoolean, JSONBTypeOf
from debusine.db.tasks import DBTask
from debusine.server.collections.base import (
    CollectionManagerInterface,
    ItemAdditionError,
)
from debusine.server.tags import (
    make_task_provided_tagset,
    make_task_required_tagset,
)
from debusine.tasks import TaskConfigError
from debusine.utils.tag_utils import DerivationRuleParser


class DebusineTaskConfigurationManager(CollectionManagerInterface):
    """Manage collection of category debusine:task-configuration."""

    COLLECTION_CATEGORY = CollectionCategory.TASK_CONFIGURATION
    VALID_BARE_DATA_CATEGORIES = frozenset(
        {BareDataCategory.TASK_CONFIGURATION}
    )

    @override
    def do_add_bare_data(
        self,
        category: BareDataCategory,
        *,
        user: User,
        workflow: WorkRequest | None = None,
        data: dict[str, Any] | None = None,
        name: str | None = None,  # noqa: U100
        replace: bool = False,
        created_at: dt.datetime | None = None,
        replaced_by: CollectionItem | None = None,
    ) -> CollectionItem:
        if data is None:
            raise ItemAdditionError(
                f"Adding to {CollectionCategory.TASK_CONFIGURATION} requires "
                f"data"
            )
        if name is not None:
            raise ItemAdditionError(
                f"Cannot use an explicit item name when adding to "
                f"{CollectionCategory.TASK_CONFIGURATION}"
            )

        config = DebusineTaskConfiguration(**data)
        name = config.name()

        if replace:
            self.remove_items_by_name(
                name=name,
                child_types=[CollectionItem.Types.BARE],
                user=user,
                workflow=workflow,
            )

        try:
            return CollectionItem.objects.create_from_bare_data(
                category,
                parent_collection=self.collection,
                name=name,
                data=data,
                created_at=created_at,
                created_by_user=user,
                created_by_workflow=workflow,
                replaced_by=replaced_by,
            )
        except IntegrityError as exc:
            raise ItemAdditionError(str(exc))

    @override
    def do_remove_item(
        self,
        item: CollectionItem,
        *,
        user: User | None = None,
        workflow: WorkRequest | None = None,
    ) -> None:
        item.removed_by_user = user
        item.removed_by_workflow = workflow
        item.removed_at = timezone.now()
        item.save()


class ConfigurationLookup:
    """Machinery to look up task configuration items."""

    def __init__(self, collection: Collection) -> None:
        """Set the collection to use for lookups."""
        self.collection = collection
        self.templates = self.collection.child_items.active().filter(
            child_type=CollectionItem.Types.BARE
        )
        self.candidates = (
            self.collection.child_items.active()
            .filter(child_type=CollectionItem.Types.BARE)
            .annotate(
                # Match via postgres only when the field is an array
                tags_provided=Case(
                    When(
                        CompareBoolean(
                            JSONBTypeOf("data__tags_provided"),
                            "=",
                            Value("array"),
                        ),
                        then="data__tags_provided",
                    ),
                    default=Value("[]"),
                    output_field=JSONField(),
                ),
                # Match via postgres only when the field is an array
                tags_required=Case(
                    When(
                        CompareBoolean(
                            JSONBTypeOf("data__tags_required"),
                            "=",
                            Value("array"),
                        ),
                        then="data__tags_required",
                    ),
                    default=Value("[]"),
                    output_field=JSONField(),
                ),
                subject=Coalesce(
                    "data__subject", Value("null"), output_field=JSONField()
                ),
                context=Coalesce(
                    "data__context", Value("null"), output_field=JSONField()
                ),
            )
        )
        self.task_tags_provided: set[str]
        self.task_tags_required: set[str]

    def set_task(self, task: DBTask[Any, Any]) -> None:
        """Set lookup constraints from a task instance."""
        self.set_task_type_name_constraints(task.TASK_TYPE, task.name)
        subject, context = task.get_task_configuration_subject_context()
        self.set_subject_context_constraints(subject, context)

        # Compute the pre-configuration scheduler tags
        config_provided = make_task_provided_tagset()
        config_required = make_task_required_tagset()
        provided, required = task.compute_scheduler_tags()
        config_provided.add_candidates(provided)
        config_provided.finalize()
        config_required.add_candidates(required)
        config_required.finalize()

        # Filter candidate entries to those matching the given tags
        self.set_tags_provided(config_provided.tags)
        self.set_tags_required(config_required.tags)

    def set_task_type_name_constraints(
        self, task_type: str, task_name: str
    ) -> None:
        """Set constraints on task type and name."""
        self.candidates = self.candidates.filter(
            data__task_type=task_type,
            data__task_name=task_name,
        )

    def set_subject_context_constraints(
        self, subject: str | None, context: str | None
    ) -> None:
        """Set constraints on subject and context fields."""
        if subject is None:
            self.candidates = self.candidates.filter(subject=None)
        else:
            self.candidates = self.candidates.filter(
                Q(subject=None) | Q(subject=subject)
            )
        if context is None:
            self.candidates = self.candidates.filter(context=None)
        else:
            self.candidates = self.candidates.filter(
                Q(context=None) | Q(context=context)
            )

    def set_tags_provided(self, tags: set[str]) -> None:
        """Set the set of task-provided tags used for filtering."""
        self.candidates = self.candidates.filter(
            tags_provided__contained_by=sorted(tags)
        )
        self.task_tags_provided = tags

    def set_tags_required(self, tags: set[str]) -> None:
        """Set the set of task-required tags used for filtering."""
        self.candidates = self.candidates.filter(
            tags_required__contained_by=sorted(tags)
        )
        self.task_tags_required = tags

    @staticmethod
    def sort_key(
        value: tuple[DebusineTaskConfiguration, int],
    ) -> tuple[str, int, int, int]:
        """Sorting key for task configuration enties."""
        # Sorting by presence of subject and context
        sc: int
        match value[0].subject, value[0].context:
            case None, None:
                sc = 0
            case None, str():
                sc = 1
            case str(), None:
                sc = 2
            case _:
                sc = 3
        return (
            # Sort by path and position
            value[0].path or "",
            value[0].position or 0,
            # Then by presence of subject and context
            sc,
            # Lastly by item ID, as a tie breaker
            value[1],
        )

    def iter_matching_entries(self) -> Generator[DebusineTaskConfiguration]:
        """Filter configuration entries and iterate them in order."""
        parser = DerivationRuleParser()
        entries: list[tuple[DebusineTaskConfiguration, int]] = []

        # Instantiate entries
        for item in self.candidates:
            try:
                parsed = DebusineTaskConfiguration.model_validate(item.data)
            except ValueError as exc:
                raise TaskConfigError(
                    f"debusine:task-configuration item {item.id}"
                    " does not validate",
                    exc,
                )
            if isinstance(parsed.tags_provided, str):
                if not parser.as_function(
                    parsed.tags_provided, "task configuration"
                )(self.task_tags_provided):
                    continue
            if isinstance(parsed.tags_required, str):
                if not parser.as_function(
                    parsed.tags_required, "task configuration"
                )(self.task_tags_required):
                    continue
            entries.append((parsed, item.id))

        # Sort items
        entries.sort(key=self.sort_key)

        yield from (e[0] for e in entries)

    def lookup_template(self, name: str) -> DebusineTaskConfiguration | None:
        """Lookup a DebusineTaskConfiguration template by name."""
        try:
            item = self.templates.get(name=f"template:{name}")
        except CollectionItem.DoesNotExist:
            return None

        try:
            return DebusineTaskConfiguration(**item.data)
        except ValueError as exc:
            raise TaskConfigError(
                "debusine:task-configuration item does not validate", exc
            )

    def lookup_templates(
        self,
        entry: DebusineTaskConfiguration,
        stack: tuple[str, ...] = (),
    ) -> Generator[DebusineTaskConfiguration]:
        """
        Generate Task Configuration objects for a list of template names.

        :param config_collection: collection to use for lookups
        :param entry: entry with the list of template names to lookup
        :param stack: template names we are currently resolving, used to detect
                      circular loops
        """
        for name in entry.use_templates:
            if name in stack:
                raise TaskConfigError(
                    f"{entry.name()}: template lookup cycle detected: "
                    + "→".join((*stack, name))
                )
            if (template_entry := self.lookup_template(name)) is None:
                raise TaskConfigError(
                    f"{entry.name()} references missing template {name!r}"
                )
            if template_entry.use_templates:
                yield from self.lookup_templates(template_entry, (*stack, name))
            yield template_entry

    def iter_items(self) -> Generator[DebusineTaskConfiguration]:
        """
        Lookup DebusineTaskConfiguration objects for an item name.

        The resulting sequence also contains referenced template elements.
        """
        for entry in self.iter_matching_entries():
            yield entry

            # Look up referenced templates
            yield from self.lookup_templates(entry)


class Changes(NamedTuple):
    """Computed changes to apply to task configuration."""

    #: Values to set if not already set
    default_values: dict[str, Any]
    #: Values to set regardless
    override_values: dict[str, Any]
    #: Extra tags to provide, with the ``workspace`` provenance
    provide_tags: set[str]
    #: Extra tags to require, with the ``workspace`` provenance
    require_tags: set[str]


def build_configuration(
    items: Iterable[DebusineTaskConfiguration],
) -> Changes:
    """Turn config items into configuration to apply to task data."""
    default_values: dict[str, Any] = {}
    override_values: dict[str, Any] = {}
    provide_tags: set[str] = set()
    require_tags: set[str] = set()
    locked_values = set()

    for config_item in items:
        # Drop all the entries referenced in `delete_values` (except
        # locked values)
        for key in config_item.delete_values:
            if key in locked_values:
                continue
            default_values.pop(key, None)
            override_values.pop(key, None)

        # Merge the default/override values in the response
        # (except locked values)
        for key, value in config_item.default_values.items():
            if key in locked_values:
                continue
            default_values[key] = value
        for key, value in config_item.override_values.items():
            if key in locked_values:
                continue
            override_values[key] = value

        # Update the set of locked values
        locked_values.update(config_item.lock_values)

        # Collected extra provided and required tags
        provide_tags.update(config_item.provide_tags)
        require_tags.update(config_item.require_tags)

    return Changes(
        default_values=default_values,
        override_values=override_values,
        provide_tags=provide_tags,
        require_tags=require_tags,
    )


def apply_configuration(
    task_data: dict[str, Any],
    config_collection: Collection,
    task: DBTask[Any, Any],
) -> Changes:
    """
    Apply task configuration to task_data.

    :param task_data: the task data dict to configure, modified in place
    :param config_collection: the task configuration collection to use
    :param task: the task to use for task configuration lookups
    :returns: the set of changes that have been applied
    """
    # Build the base setup for looking up configuration entries
    lookup = ConfigurationLookup(config_collection)
    lookup.set_task(task)

    # Generate a list of matching items and their used template entries
    config_items = lookup.iter_items()

    # Turn items into a configuration to apply
    changes = build_configuration(config_items)

    # Apply default values (add missing values, but also replace explicit
    # None values)
    for k, v in changes.default_values.items():
        if task_data.get(k) is None:
            task_data[k] = v

    # Apply overrides
    task_data.update(changes.override_values)

    # Store the ID of the task configuration collection that was used
    task_data["task_configuration"] = config_collection.id

    return changes



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to list tokens. Allows filtering."""

from typing import Any, override

from django.core.management import CommandParser

from debusine.db.models import Token
from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.server.management.management_utils import Tokens


class Command(DebusineBaseCommand):
    """Command to list tokens."""

    help = "List tokens. See the options to filter which tokens will be listed"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument('--token', help='List only this token')
        parser.add_argument('--username', help='List tokens owner by USERNAME')
        parser.add_argument(
            '--yaml', action="store_true", help="Machine readable YAML output"
        )

    @override
    def handle(self, *args: Any, **options: Any) -> None:
        """List the tokens."""
        tokens = Token.objects.get_tokens(
            username=options["username"],
            key=options["token"],
            include_expired=True,
        )
        Tokens(options["yaml"]).print(tokens, self.stdout)
        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the workspace views."""

import datetime as dt
import json
from typing import Any, ClassVar, NamedTuple, override
from unittest import mock

import lxml
import lxml.objectify
from django.contrib import messages
from django.db.models import Max
from django.template.response import SimpleTemplateResponse
from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from debusine.artifacts.models import (
    BareDataCategory,
    CollectionCategory,
    DebusineTaskConfiguration,
    TaskTypes,
)
from debusine.db.models import Collection, Scope, WorkRequest, Workspace
from debusine.db.playground import scenarios
from debusine.test.django import (
    AllowAll,
    TestCase,
    TestResponseType,
    override_permission,
)
from debusine.web.views.tests.utils import ViewTestMixin


class WorkspaceDetailViewTests(ViewTestMixin, TestCase):
    """Tests for WorkspaceDetailView class."""

    scenario = scenarios.DefaultScopeUser()

    workspace_public: ClassVar[Workspace]
    workspace_private: ClassVar[Workspace]
    collection: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Set up a database layout for views."""
        super().setUpTestData()
        cls.workspace_public = cls.playground.create_workspace(
            name="Public", public=True
        )
        cls.workspace_private = cls.playground.create_workspace(name="Private")

        cls.collection = cls.playground.create_collection(
            "Public",
            CollectionCategory.WORKFLOW_INTERNAL,
            workspace=cls.workspace_public,
        )

    def test_detail(self) -> None:
        """Test workspace detail view."""
        response = self.client.get(self.workspace_public.get_absolute_url())
        tree = self.assertResponseHTML(response)
        details = self.assertHasElement(
            tree, "//table[@id='workspace-details']"
        )
        self.assertHasElement(details, ".//tr[@id='workspace-details-public']")
        self.assertHasElement(
            details, ".//tr[@id='workspace-details-default-expiration']"
        )

    @override_settings(
        DEBUSINE_DEBIAN_ARCHIVE_URL_SCHEME="https",
        DEBUSINE_DEBIAN_ARCHIVE_PRIMARY_FQDN="deb.example.org",
    )
    def test_debian_archive_url(self) -> None:
        Collection.objects.create(
            workspace=self.workspace_public,
            category=CollectionCategory.ARCHIVE,
            name="_",
        )
        response = self.client.get(self.workspace_public.get_absolute_url())
        tree = self.assertResponseHTML(response)
        a = self.assertHasElement(
            tree,
            "//table[@id='workspace-details']"
            "//tr[@id='workspace-debian-archive']/td/a",
        )
        url = (
            f"https://deb.example.org/{self.scenario.scope.name}/"
            f"{self.workspace_public.name}/"
        )
        self.assertEqual(a.get("href"), url)
        self.assertTextContentEqual(a, url)

    @override_settings(DEBUSINE_DEBIAN_ARCHIVE_PRIMARY_FQDN="deb.example.org")
    def test_debian_archive_url_archive_not_present(self) -> None:
        Collection.objects.filter(
            workspace=self.workspace_public,
            category=CollectionCategory.ARCHIVE,
        ).delete()
        response = self.client.get(self.workspace_public.get_absolute_url())
        tree = self.assertResponseHTML(response)
        self.assertFalse(
            tree.xpath(
                "//table[@id='workspace-details']"
                "//tr[@id='workspace-debian-archive']"
            )
        )

    def test_permissions(self) -> None:
        self.assertSetsCurrentWorkspace(
            self.workspace_public,
            self.workspace_public.get_absolute_url(),
        )

    def test_no_expiration(self) -> None:
        """No expiration is shown on non-expiring workspaces."""
        self.client.force_login(self.scenario.user)
        response = self.client.get(self.workspace_public.get_absolute_url())
        tree = self.assertResponseHTML(response)
        details = self.assertHasElement(
            tree, "//table[@id='workspace-details']"
        )
        tr = self.assertHasElement(
            details, "//tr[@id='workspace-details-expiration']"
        )
        self.assertTextContentEqual(tr.th, "Expires")
        self.assertTextContentEqual(tr.td, "Never")

    def test_expiration(self) -> None:
        """Expiration is shown on expiring workspaces."""
        self.workspace_public.expiration_delay = dt.timedelta(days=7)
        self.workspace_public.save()

        self.client.force_login(self.scenario.user)
        response = self.client.get(self.workspace_public.get_absolute_url())
        tree = self.assertResponseHTML(response)
        details = self.assertHasElement(
            tree, "//table[@id='workspace-details']"
        )
        tr = self.assertHasElement(
            details, "//tr[@id='workspace-details-expiration']"
        )
        self.assertTextContentEqual(tr.th, "Expires")
        self.assertTextContentEqual(
            tr.td,
            (self.workspace_public.created_at + dt.timedelta(days=7)).strftime(
                "%Y-%m-%d"
            ),
        )

    def test_detail_parents(self) -> None:
        """Parent workspaces are shown when they exist."""
        workspace = self.playground.create_workspace(name="base", public=True)
        workspace.set_inheritance(
            [self.workspace_private, self.workspace_public]
        )

        response = self.client.get(workspace.get_absolute_url())
        tree = self.assertResponseHTML(response)
        details = self.assertHasElement(
            tree, "//table[@id='workspace-details']"
        )
        tr = self.assertHasElement(
            details, "*/tr[@id='workspace-details-parents']", dump_on_error=True
        )
        self.assertTextContentEqual(tr.th, "Inherits from")
        self.assertTextContentEqual(tr.td, str(self.workspace_public))

    def test_can_configure(self) -> None:
        """User is logged in and can create work_request."""
        self.client.force_login(self.scenario.user)
        with (
            override_permission(Workspace, "can_configure", AllowAll),
        ):
            response = self.client.get(self.workspace_public.get_absolute_url())
        tree = self.assertResponseHTML(response)
        self.assertHasElement(tree, "//div[@id='workspace-actions']")

    def test_private(self) -> None:
        """Unauthenticated detail view cannot access private workspaces."""
        response = self.client.get(self.workspace_private.get_absolute_url())
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.client.force_login(self.playground.get_default_user())
        response = self.client.get(self.workspace_private.get_absolute_url())
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.playground.create_group_role(
            self.workspace_private,
            Workspace.Roles.OWNER,
            users=[self.playground.get_default_user()],
        )
        response = self.client.get(self.workspace_private.get_absolute_url())
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_private_external_user_token(self) -> None:
        """External user tokens do not grant access to private workspaces."""
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST, workspace=self.workspace_private
        )
        token = self.playground.create_external_user_token()
        collection.grant_role(
            token, Collection.Roles.VIEWER, created_by=self.scenario.user
        )

        response = self.client.get(
            self.workspace_private.get_absolute_url(), HTTP_TOKEN=token.key
        )

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_name_duplicate(self) -> None:
        """Homonymous workspace in different scope does not trigger a dup."""
        scope = Scope.objects.create(name="tests")
        self.playground.create_workspace(
            name="Public", public=True, scope=scope
        )
        response = self.client.get(self.workspace_public.get_absolute_url())
        self.assertResponseHTML(response)

    def test_workflowtemplate_list_empty(self) -> None:
        """Test showing a workspace with no workflow templates."""
        response = self.client.get(self.workspace_public.get_absolute_url())
        tree = self.assertResponseHTML(response)
        div = self.assertHasElement(tree, "//div[@id='workflow-templates']")
        self.assertTextContentEqual(
            div, "No workflow templates configured for this workspace."
        )

    def test_workflowtemplate(self) -> None:
        """Test showing a workspace with one workflow templates."""
        template = self.playground.create_workflow_template(
            name="noop-template",
            task_name="noop",
            workspace=self.workspace_public,
        )

        # 3 completed workflow
        for _ in range(3):
            self.playground.create_workflow(
                task_name="noop", result=WorkRequest.Results.SUCCESS
            )

        # 2 running workflows
        for _ in range(2):
            self.playground.create_workflow(task_name="noop", mark_running=True)

        # 1 needs input workflow
        workflow_needs_input = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        workflow_needs_input.workflow_runtime_status = (
            WorkRequest.RuntimeStatuses.NEEDS_INPUT
        )
        workflow_needs_input.save()

        response = self.client.get(self.workspace_public.get_absolute_url())
        tree = self.assertResponseHTML(response)

        self.assertTrue(response.context["workflow_templates"].ordered)

        div = self.assertHasElement(tree, "//div[@id='workflow-templates']")
        workflow_templates = div.table

        self.assertEqual(
            workflow_templates.attrib["id"], "workflow-templates-list"
        )

        self.assertEqual(
            workflow_templates.tbody.tr[0].td[0].a, "noop-template"
        )
        self.assertEqual(
            workflow_templates.tbody.tr[0].td[0].a.attrib["href"],
            template.get_absolute_url(),
        )
        self.assertEqual(workflow_templates.tbody.tr[0].td[1].a, "noop")
        self.assertEqual(
            workflow_templates.tbody.tr[0].td[1].a.attrib["href"],
            "https://freexian-team.pages.debian.net/debusine/"
            "reference/workflows/specs/noop.html",
        )

    def test_workflowtemplate_links(self) -> None:
        """Test links to workflow tables."""
        workflow_list_base_url = reverse(
            "workspaces:workflows:list",
            kwargs={"wname": self.workspace_public.name},
        )
        self.playground.create_workflow_template(
            name="noop-template",
            task_name="noop",
            workspace=self.workspace_public,
        )

        # 1 completed workflow
        self.playground.create_workflow(
            task_name="noop", result=WorkRequest.Results.SUCCESS
        )

        response = self.client.get(self.workspace_public.get_absolute_url())
        tree = self.assertResponseHTML(response)

        div = self.assertHasElement(tree, "//div[@id='workflow-templates']")
        tr = div.table.tbody.tr[0]

        # Running workflows
        url = tr.td[2].a.get("href")
        assert url is not None
        self.assertIn(workflow_list_base_url, url)
        response = self.client.get(url)
        paginator = response.context["paginator"]
        self.assertEqual(
            paginator.table.filters["workflow_templates"].value, "noop-template"
        )
        self.assertEqual(
            paginator.table.filters["statuses"].value, ["running__any"]
        )

        # Input needed workflows
        url = tr.td[3].a.get("href")
        assert url is not None
        self.assertIn(workflow_list_base_url, url)
        response = self.client.get(url)
        paginator = response.context["paginator"]
        self.assertEqual(
            paginator.table.filters["workflow_templates"].value, "noop-template"
        )
        self.assertEqual(
            paginator.table.filters["statuses"].value, ["running__needs_input"]
        )

        # Completed workflows
        url = tr.td[4].a.get("href")
        assert url is not None
        self.assertIn(workflow_list_base_url, url)
        response = self.client.get(url)
        paginator = response.context["paginator"]
        self.assertEqual(
            paginator.table.filters["workflow_templates"].value, "noop-template"
        )
        self.assertEqual(
            paginator.table.filters["statuses"].value, ["completed"]
        )


class WorkspaceUpdateViewTests(ViewTestMixin, TestCase):
    """Tests for WorkspaceUpdateView class."""

    scenario = scenarios.DefaultContext()

    def test_permissions(self) -> None:
        """Test permissions."""
        self.assertSetsCurrentWorkspace(
            self.scenario.workspace,
            self.scenario.workspace.get_absolute_url_configure(),
        )
        self.assertEnforcesPermission(
            self.scenario.workspace.can_configure,
            self.scenario.workspace.get_absolute_url_configure(),
            "get_context_data",
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_configure(self) -> None:
        """Test changing workspace configuration."""
        ws = self.scenario.workspace
        self.playground.create_group_role(
            ws, Workspace.Roles.OWNER, users=[self.scenario.user]
        )
        self.client.force_login(self.scenario.user)

        response = self.client.post(
            ws.get_absolute_url_configure(),
            {
                "public": False,
                "expiration_delay": 7,
                "default_expiration_delay": 14,
            },
        )
        self.assertRedirects(response, ws.get_absolute_url())

        ws.refresh_from_db()

        self.assertFalse(ws.public)
        self.assertEqual(ws.expiration_delay, dt.timedelta(days=7))
        self.assertEqual(ws.default_expiration_delay, dt.timedelta(days=14))

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_initial_form(self) -> None:
        """Get request to ensure the form is displayed."""
        ws = self.scenario.workspace
        self.client.force_login(self.scenario.user)

        title_text = "Configure"
        response = self.client.get(ws.get_absolute_url_configure())
        tree = self.assertResponseHTML(response)
        title = self.assertHasElement(tree, "//head//title")
        self.assertTextContentEqual(title, f"debusine - System - {title_text}")
        h1 = self.assertHasElement(tree, "//body//h1")
        self.assertTextContentEqual(h1, title_text)

        assert isinstance(response, SimpleTemplateResponse)
        assert response.context_data is not None
        form = response.context_data["form"]
        self.assertTrue(form["public"].initial)
        self.assertIsNone(form["expiration_delay"].initial)
        self.assertEqual(
            form["default_expiration_delay"].initial, dt.timedelta(0)
        )


class TaskConfigInspectorViewTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`TaskConfigInspectorView` class."""

    scenario = scenarios.DefaultContext()
    collection: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Set up a database layout for views."""
        super().setUpTestData()
        cls.collection = cls.scenario.default_task_configuration_collection

    @classmethod
    def add_config(cls, entry: DebusineTaskConfiguration) -> None:
        """Add a config entry to config_collection."""
        cls.collection.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=cls.scenario.user,
            data=entry,
        )

    def get(
        self,
        collections: list[Collection] | None = None,
        **kwargs: str,
    ) -> TestResponseType:
        workspace = self.scenario.workspace
        if collections is None:
            collections = [self.collection]
        url = reverse(
            "workspaces:task_configuration_inspector",
            kwargs={"wname": workspace.name},
        )

        with mock.patch(
            "debusine.db.models.workspaces."
            "Workspace.list_accessible_collections",
            return_value=collections,
        ):
            return self.client.get(url, kwargs)

    def assertHasInputContainer(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> lxml.objectify.ObjectifiedElement:
        """Return the main user input container element."""
        return self.assertHasElement(tree, "//div[@id='user-input']")

    def assertHasCollectionSelector(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> lxml.objectify.ObjectifiedElement:
        """Return the collection selector root element."""
        input_container = self.assertHasInputContainer(tree)
        return self.assertHasElement(
            input_container, "//div[@id='collection-selector']"
        )

    def assertCollectionSelected(
        self,
        tree: lxml.objectify.ObjectifiedElement,
        collection: Collection,
        can_reselect: bool = False,
        is_empty: bool = False,
    ) -> None:
        """Ensure the given collection is currently selected."""
        input_container = self.assertHasInputContainer(tree)

        # Ensure the collection selector is not shown
        self.assertFalse(
            input_container.xpath("//div[@id='collection-selector']")
        )

        input_header = self.assertHasElement(
            input_container, "div[contains(@class, 'card-header')]"
        )
        input_title = self.assertHasElement(input_header, "*/h2")
        self.assertTextContentEqual(
            input_title,
            f"Collection {collection.name} in {collection.workspace}",
        )

        if can_reselect:
            reselect = self.assertHasElement(input_header, "a[@href='.']")
            self.assertEqual(reselect.get("title"), "select another collection")
        else:
            self.assertFalse(input_header.xpath("a[@href='.']"))

        input_body = self.assertHasElement(
            input_container, "div[contains(@class, 'card-body')]"
        )
        if is_empty:
            self.assertTextContentEqual(input_body, "This collection is empty")
        else:
            selected_collection = self.assertHasElement(
                input_body, "form/input[@name='collection']"
            )
            self.assertEqual(
                selected_collection.get("value"), str(collection.id)
            )

    def test_title(self) -> None:
        expected_title = "Task configuration inspector"

        response = self.get()
        tree = self.assertResponseHTML(response)
        title = self.assertHasElement(tree, "//head//title")
        self.assertTextContentEqual(
            title, f"debusine - System - {expected_title}"
        )
        h1 = self.assertHasElement(tree, "//body//h1")
        self.assertTextContentEqual(h1, expected_title)

    def test_collection_selector(self) -> None:
        ws_other = self.playground.create_workspace(name="Other", public=True)
        coll_other = self.playground.create_collection(
            "default", CollectionCategory.TASK_CONFIGURATION, workspace=ws_other
        )

        response = self.get(collections=[self.collection, coll_other])
        tree = self.assertResponseHTML(response)
        div = self.assertHasCollectionSelector(tree)
        self.assertEqual(
            div.a[0].get("href"), f"?collection={self.collection.pk}"
        )
        self.assertTextContentEqual(
            div.a[0], f"{self.collection.name} in {self.collection.workspace}"
        )
        self.assertEqual(div.a[1].get("href"), f"?collection={coll_other.pk}")
        self.assertTextContentEqual(
            div.a[1], f"{coll_other.name} in {coll_other.workspace}"
        )

    def test_collection_selected(self) -> None:
        ws_other = self.playground.create_workspace(name="Other", public=True)
        coll_other = self.playground.create_collection(
            "default", CollectionCategory.TASK_CONFIGURATION, workspace=ws_other
        )

        response = self.get(
            collections=[self.collection, coll_other],
            collection=str(coll_other.id),
        )
        tree = self.assertResponseHTML(response)
        self.assertCollectionSelected(
            tree, coll_other, can_reselect=True, is_empty=True
        )

    def test_collection_autoselect(self) -> None:
        response = self.get(collections=[self.collection])
        tree = self.assertResponseHTML(response)
        self.assertCollectionSelected(
            tree, self.collection, can_reselect=False, is_empty=True
        )

    def test_collection_lookup(self) -> None:
        item = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.add_config(item)
        response = self.get(
            collections=[self.collection], task=f"{TaskTypes.WORKER}:noop"
        )
        tree = self.assertResponseHTML(response)
        self.assertCollectionSelected(
            tree, self.collection, can_reselect=False, is_empty=False
        )

        results = self.assertHasElement(tree, "//div[@id='results']")

        default_values = self.assertHasElement(
            results, "div[@id='default-values']"
        )
        self.assertYAMLContentEqual(default_values.div[1], {})
        override_values = self.assertHasElement(
            results, "div[@id='override-values']"
        )
        self.assertYAMLContentEqual(override_values.div[1], {})
        configuration_items = self.assertHasElement(
            results, "div[@id='configuration-items']"
        )
        self.assertTextContentEqual(
            configuration_items.div[0].div[0], item.name()
        )
        self.assertYAMLContentEqual(
            configuration_items.div[0].div[1], item.model_dump()
        )

    def test_collection_tags_provided(self) -> None:
        item = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.add_config(item)
        with self.spy(
            "debusine.server.collections.debusine_task_configuration"
            ".ConfigurationLookup.set_tags_provided"
        ) as set_tp:
            self.assertResponseHTML(
                self.get(
                    collections=[self.collection],
                    task=f"{TaskTypes.WORKER}:noop",
                    tags_provided="a b c",
                )
            )
        set_tp.assert_called_once()
        self.assertEqual(set_tp.call_args.args[1], {"a", "b", "c"})

    def test_collection_tags_required(self) -> None:
        item = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.add_config(item)
        with self.spy(
            "debusine.server.collections.debusine_task_configuration"
            ".ConfigurationLookup.set_tags_required"
        ) as set_tr:
            self.assertResponseHTML(
                self.get(
                    collections=[self.collection],
                    task=f"{TaskTypes.WORKER}:noop",
                    tags_required="a b c",
                )
            )
        set_tr.assert_called_once()
        self.assertEqual(set_tr.call_args.args[1], {"a", "b", "c"})

    def test_form_invalid(self) -> None:
        item = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.add_config(item)
        response = self.get(
            collections=[self.collection], task=f"{TaskTypes.WORKER}:invalid"
        )
        tree = self.assertResponseHTML(response)
        self.assertCollectionSelected(
            tree, self.collection, can_reselect=False, is_empty=False
        )

        self.assertFalse(tree.xpath("//div[@id='results']"))

    def test_invalid_collection_parameter(self) -> None:
        """Invalid collection parameter returns correct status codes."""
        test_cases = [
            ("nonexistent", 999999999, status.HTTP_404_NOT_FOUND),
            ("non_numeric", "abc", status.HTTP_400_BAD_REQUEST),
            ("empty", "", status.HTTP_400_BAD_REQUEST),
            ("overflow", 10**100, status.HTTP_404_NOT_FOUND),
        ]

        for label, value, expected_status in test_cases:
            with self.subTest(label=label, value=value):
                response = self.get(collection=str(value))
                self.assertEqual(response.status_code, expected_status)


class WorkspaceUpdateInheritanceViewTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`WorkspaceUpdateInheritanceView` class."""

    scenario = scenarios.DefaultContext()
    ws1: ClassVar[Workspace]
    ws2: ClassVar[Workspace]

    class PostForm(NamedTuple):
        pk: int
        order: int
        delete: bool = False

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Set up a database layout for views."""
        super().setUpTestData()
        cls.ws1 = cls.playground.create_workspace(name="ws1", public=True)
        cls.ws2 = cls.playground.create_workspace(name="ws2", public=True)

    def get(self, **kwargs: str) -> TestResponseType:
        workspace = self.scenario.workspace
        url = reverse(
            "workspaces:update_inheritance",
            kwargs={"wname": workspace.name},
        )
        return self.client.get(url, kwargs)

    def post(
        self,
        items: list[PostForm | Workspace] | None = None,
        initial_forms: int | None = None,
        save: str = "continue",
    ) -> TestResponseType:
        """
        Make a POST request.

        :param items: list of data for each individual formset form. If omitted
            it uses the empty list. If passed a Workspace it uses its ID and
            infers the rest.
        :param initial_forms: number of initial forms in the formset management
            form. If omitted the length of items is used.
        :param save: value of the ``save`` field.
        """
        post_data: dict[str, Any] = {}
        form_count = 0
        for item in items or []:
            form: WorkspaceUpdateInheritanceViewTests.PostForm
            if isinstance(item, Workspace):
                form = self.PostForm(item.pk, form_count, False)
            else:
                form = item
            post_data[f"form-{form_count}-parent"] = form.pk
            post_data[f"form-{form_count}-ORDER"] = form.order
            if form.delete:
                post_data[f"form-{form_count}-DELETE"] = "true"
            form_count += 1
        post_data[f"form-{form_count}-parent"] = ""
        post_data[f"form-{form_count}-ORDER"] = ""
        if initial_forms is None:
            initial_forms = len(items) if items is not None else 0
        post_data["form-TOTAL_FORMS"] = max(form_count + 1, initial_forms)
        post_data["form-INITIAL_FORMS"] = initial_forms
        post_data["form-MIN_NUM_FORMS"] = 0
        post_data["form-MAX_NUM_FORMS"] = 1000
        post_data["save"] = save
        url = reverse(
            "workspaces:update_inheritance",
            kwargs={"wname": self.scenario.workspace.name},
        )
        return self.client.post(url, data=post_data)

    def assertParents(self, workspace: Workspace, names: list[str]) -> None:
        """Ensure the inheritance chain matches the listed workspace names."""
        self.assertQuerySetEqual(
            workspace.chain_parents.order_by("order").values_list(
                "parent__name", flat=True
            ),
            names,
        )

    def assertHasMessage(
        self, response: TestResponseType, level: int, text: str
    ) -> None:
        """Ensure the response sent the given Django message."""
        msgs = messages.get_messages(response.wsgi_request)
        assert [(m.level, m.message) for m in msgs] == [(level, text)]

    def assertHasNoMessages(self, response: TestResponseType) -> None:
        """Ensure the response did not send any Django message."""
        msgs = messages.get_messages(response.wsgi_request)
        assert [m.message for m in msgs] == []

    def test_permissions(self) -> None:
        """Test permissions."""
        self.assertSetsCurrentWorkspace(
            self.scenario.workspace,
            self.scenario.workspace.get_absolute_url_configure_inheritance(),
        )
        self.assertSetsCurrentWorkspace(
            self.scenario.workspace,
            self.scenario.workspace.get_absolute_url_configure_inheritance(),
            method="post",
        )
        self.assertEnforcesPermission(
            self.scenario.workspace.can_configure,
            self.scenario.workspace.get_absolute_url_configure_inheritance(),
            "get_object",
        )
        self.assertEnforcesPermission(
            self.scenario.workspace.can_configure,
            self.scenario.workspace.get_absolute_url_configure_inheritance(),
            "get_object",
            method="post",
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_title(self) -> None:
        expected_title = "Configure inheritance"

        response = self.get()
        tree = self.assertResponseHTML(response)
        title = self.assertHasElement(tree, "//head//title")
        self.assertTextContentEqual(
            title, f"debusine - System - {expected_title}"
        )
        h1 = self.assertHasElement(tree, "//body//h1")
        self.assertTextContentEqual(h1, expected_title)

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_get(self) -> None:
        response = self.get()
        self.assertEqual(
            json.loads(response.context["available"]),
            [[self.ws1.pk, str(self.ws1)], [self.ws2.pk, str(self.ws2)]],
        )
        self.assertEqual(json.loads(response.context["current"]), [])

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_get_with_current_chain(self) -> None:
        self.scenario.workspace.set_inheritance([self.ws2, self.ws1])
        response = self.get()
        self.assertEqual(
            json.loads(response.context["available"]),
            [[self.ws1.pk, str(self.ws1)], [self.ws2.pk, str(self.ws2)]],
        )
        self.assertEqual(
            json.loads(response.context["current"]),
            [[self.ws2.pk, str(self.ws2)], [self.ws1.pk, str(self.ws1)]],
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_empty(self) -> None:
        response = self.post()
        self.assertRedirects(
            response,
            self.scenario.workspace.get_absolute_url_configure_inheritance(),
        )
        self.scenario.workspace.refresh_from_db()
        self.assertParents(self.scenario.workspace, [])
        self.assertHasMessage(
            response, messages.SUCCESS, "Inheritance chain saved"
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_add(self) -> None:
        self.scenario.workspace.set_inheritance([self.ws2])
        response = self.post([self.ws1, self.ws2])
        self.assertRedirects(
            response,
            self.scenario.workspace.get_absolute_url_configure_inheritance(),
        )
        self.scenario.workspace.refresh_from_db()
        self.assertParents(self.scenario.workspace, ["ws1", "ws2"])
        self.assertHasMessage(
            response, messages.SUCCESS, "Inheritance chain saved"
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_add_save_done(self) -> None:
        self.scenario.workspace.set_inheritance([self.ws2, self.ws1])
        response = self.post([self.ws1, self.ws2], save="done")
        self.assertRedirects(
            response,
            self.scenario.workspace.get_absolute_url(),
        )
        self.scenario.workspace.refresh_from_db()
        self.assertParents(self.scenario.workspace, ["ws1", "ws2"])
        self.assertHasMessage(
            response, messages.SUCCESS, "Inheritance chain saved"
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_delete(self) -> None:
        self.scenario.workspace.set_inheritance([self.ws2, self.ws1])
        response = self.post([self.PostForm(self.ws1.pk, 10, True), self.ws2])
        self.assertRedirects(
            response,
            self.scenario.workspace.get_absolute_url_configure_inheritance(),
        )
        self.scenario.workspace.refresh_from_db()
        self.assertParents(self.scenario.workspace, ["ws2"])
        self.assertHasMessage(
            response, messages.SUCCESS, "Inheritance chain saved"
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_invalid_empty(self) -> None:
        self.scenario.workspace.set_inheritance([self.ws2, self.ws1])
        response = self.post(initial_forms=2)
        self.assertResponseHTML(response)
        self.assertEqual(
            response.context["form"].errors,
            [
                {"parent": ["This field is required."]},
                {"parent": ["This field is required."]},
            ],
        )
        self.assertHasNoMessages(response)

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_invalid_id(self) -> None:
        invalid_id = Workspace.objects.aggregate(id_max=Max("id"))["id_max"] + 1
        response = self.post([self.PostForm(invalid_id, 1)])
        self.assertResponseHTML(response)
        self.assertEqual(
            response.context["form"].errors,
            [
                {
                    "parent": [
                        "Select a valid choice."
                        f" {invalid_id} is not one of the available choices."
                    ]
                },
                {},
            ],
        )
        self.assertHasNoMessages(response)

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_invalid_duplicate(self) -> None:
        response = self.post([self.ws1, self.ws1])
        self.assertRedirects(
            response,
            self.scenario.workspace.get_absolute_url_configure_inheritance(),
        )
        self.assertHasMessage(
            response,
            messages.ERROR,
            "duplicate workspace 'ws1' in inheritance chain",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the ``update_suite`` workflow."""

from typing import ClassVar, override

import pydantic_core
from django.db import connection

from debusine.artifacts.models import CollectionCategory, TaskTypes
from debusine.db.models import Collection, CollectionItem, WorkRequest
from debusine.server.workflows import UpdateSuiteWorkflow
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.models import UpdateSuiteData
from debusine.server.workflows.tests.helpers import WorkflowTestCase
from debusine.signing.models import SigningMode
from debusine.tasks.models import BaseDynamicTaskData


class UpdateSuiteWorkflowTests(WorkflowTestCase[UpdateSuiteWorkflow]):
    """Unit tests for :py:class:`UpdateSuiteWorkflow`."""

    SAMPLE_TASK_DATA = {"suite_collection": f"sid@{CollectionCategory.SUITE}"}

    sid: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.sid = cls.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )

    def create_release_item(
        self, suite: Collection, work_request: WorkRequest
    ) -> CollectionItem:
        """Create a minimal debian:repository-index item in the suite."""
        return CollectionItem.objects.create_from_artifact(
            self.playground.create_repository_index(
                "Release", work_request=work_request
            ),
            parent_collection=suite,
            name="index:Release",
            data={"path": "Release"},
            created_by_user=self.playground.get_default_user(),
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            BaseDynamicTaskData(subject="sid"),
        )

    def test_populate(self) -> None:
        suite = self.sid
        workflow = self.playground.create_workflow(
            task_name="update_suite",
            task_data=UpdateSuiteData(
                suite_collection=f"sid@{CollectionCategory.SUITE}"
            ),
            mark_running=True,
        )

        self.assertTrue(orchestrate_workflow(workflow))

        with connection.cursor() as cursor:
            cursor.execute("SELECT CURRENT_TIMESTAMP")
            [now] = cursor.fetchone()
        self.assertQuerySetEqual(
            workflow.children.values_list(
                "task_type",
                "task_name",
                "task_data",
                "workflow_data_json",
                "status",
            ),
            [
                (
                    TaskTypes.SERVER,
                    "generatesuiteindexes",
                    {
                        "suite_collection": f"sid@{CollectionCategory.SUITE}",
                        "generate_at": pydantic_core.to_jsonable_python(now),
                    },
                    {
                        "display_name": "Generate indexes for sid",
                        "step": "generate-suite-indexes-sid",
                    },
                    WorkRequest.Statuses.PENDING,
                ),
                (
                    TaskTypes.SIGNING,
                    "signrepositoryindex",
                    {
                        "suite_collection": f"sid@{CollectionCategory.SUITE}",
                        "unsigned": (
                            "internal@collections/name:generated-release-sid"
                        ),
                        "mode": SigningMode.DETACHED,
                        "signed_name": "Release.gpg",
                    },
                    {
                        "display_name": "Create signed Release.gpg for sid",
                        "step": "sign-detached-sid",
                    },
                    WorkRequest.Statuses.BLOCKED,
                ),
                (
                    TaskTypes.SIGNING,
                    "signrepositoryindex",
                    {
                        "suite_collection": f"sid@{CollectionCategory.SUITE}",
                        "unsigned": (
                            "internal@collections/name:generated-release-sid"
                        ),
                        "mode": SigningMode.CLEAR,
                        "signed_name": "InRelease",
                    },
                    {
                        "display_name": "Create signed InRelease for sid",
                        "step": "sign-clear-sid",
                    },
                    WorkRequest.Statuses.BLOCKED,
                ),
            ],
        )
        self.assertQuerySetEqual(
            workflow.children.values_list(
                "workflow_data_json__step",
                "dependencies__workflow_data_json__step",
            ),
            [
                ("generate-suite-indexes-sid", None),
                ("sign-detached-sid", "generate-suite-indexes-sid"),
                ("sign-clear-sid", "generate-suite-indexes-sid"),
                ("sign-clear-sid", "sign-detached-sid"),
            ],
            ordered=False,
        )

        # Completing each step allows the next to proceed.
        generate_suite_indexes = workflow.children.get(
            workflow_data_json__step="generate-suite-indexes-sid"
        )
        release_item = self.create_release_item(suite, generate_suite_indexes)

        # Resolve the promise in the internal collection
        # FIXME: is there a better way to do this in tests?
        assert workflow.internal_collection is not None
        assert release_item.artifact is not None
        workflow.internal_collection.manager.add_artifact(
            release_item.artifact,
            user=release_item.created_by_user,
            workflow=workflow,
            name="generated-release-sid",
            variables=release_item.data,
            replace=True,
        )

        self.assertQuerySetEqual(
            suite.child_items.values_list("artifact", "data__path"),
            [(release_item.artifact_id, "Release")],
            ordered=False,
        )
        self.playground.advance_work_request(
            generate_suite_indexes, result=WorkRequest.Results.SUCCESS
        )

        sign_detached = workflow.children.get(
            workflow_data_json__step="sign-detached-sid"
        )
        self.assertEqual(sign_detached.status, WorkRequest.Statuses.PENDING)
        sign_clear = workflow.children.get(
            workflow_data_json__step="sign-clear-sid"
        )
        self.assertEqual(sign_clear.status, WorkRequest.Statuses.BLOCKED)

        release_gpg = self.playground.create_repository_index(
            "Release.gpg", work_request=sign_detached
        )
        self.playground.advance_work_request(
            sign_detached, result=WorkRequest.Results.SUCCESS
        )

        self.assertQuerySetEqual(
            suite.child_items.values_list("artifact", "data__path"),
            [
                (release_item.artifact_id, "Release"),
                (release_gpg.id, "Release.gpg"),
            ],
            ordered=False,
        )
        sign_clear.refresh_from_db()
        self.assertEqual(sign_clear.status, WorkRequest.Statuses.PENDING)

        inrelease = self.playground.create_repository_index(
            "InRelease", work_request=sign_clear
        )
        self.playground.advance_work_request(
            sign_clear, result=WorkRequest.Results.SUCCESS
        )

        self.assertQuerySetEqual(
            suite.child_items.values_list("artifact", "data__path"),
            [
                (release_item.artifact_id, "Release"),
                (release_gpg.id, "Release.gpg"),
                (inrelease.id, "InRelease"),
            ],
            ordered=False,
        )
